# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for upload_testcase."""

import datetime
import io
import os
import unittest

import flask

from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers import upload_testcase
from libs import helpers

DATA_DIRECTORY = os.path.join(os.path.dirname(__file__), 'upload_testcase_data')


@test_utils.with_cloud_emulators('datastore')
class FindFuzzTargetTest(unittest.TestCase):
  """Tests for find_fuzz_target."""

  def setUp(self):
    test_helpers.patch_environ(self)

    data_types.FuzzTarget(
        engine='libFuzzer', project='test-project', binary='binary').put()

    data_types.FuzzTarget(
        engine='libFuzzer', project='proj', binary='binary').put()

  def test_without_project_prefix(self):
    """Test find_fuzz_target with a target_name that isn't prefixed with the
    project."""
    data_types.Job(name='job', environment_string='PROJECT_NAME = proj\n').put()
    self.assertEqual(('libFuzzer_proj_binary', 'binary'),
                     upload_testcase.find_fuzz_target('libFuzzer', 'binary',
                                                      'job'))

  def test_with_project_prefix(self):
    """Test find_fuzz_target with a target_name that is prefixed with the
    project."""
    data_types.Job(name='job', environment_string='PROJECT_NAME = proj\n').put()
    self.assertEqual(('libFuzzer_proj_binary', 'binary'),
                     upload_testcase.find_fuzz_target('libFuzzer',
                                                      'proj_binary', 'job'))

  def test_with_main_project(self):
    """Test find_fuzz_target with a target in the main project."""
    data_types.Job(name='job', environment_string='').put()
    self.assertEqual(('libFuzzer_binary', 'binary'),
                     upload_testcase.find_fuzz_target('libFuzzer', 'binary',
                                                      'job'))

  def test_not_found(self):
    """Test target not found."""
    data_types.Job(name='job', environment_string='').put()
    with self.assertRaises(helpers.EarlyExitException):
      self.assertEqual((None, None),
                       upload_testcase.find_fuzz_target('libFuzzer', 'notfound',
                                                        'job'))


# pylint: disable=protected-access
@test_utils.with_cloud_emulators('datastore')
class UploadOAuthTest(unittest.TestCase):
  """OAuth upload tests."""

  def setUp(self):
    self.maxDiff = None  # pylint: disable=invalid-name
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.blobs.get_blob_info',
        'clusterfuzz._internal.google_cloud_utils.blobs.write_blob',
        'libs.access.has_access',
        'libs.helpers.get_user_email',
        'clusterfuzz._internal.base.utils.current_date_time',
        'clusterfuzz._internal.base.utils.utcnow',
    ])

    self.mock.get_blob_info.return_value.filename = 'input'
    self.mock.get_blob_info.return_value.key.return_value = 'blob_key'
    self.mock.write_blob.return_value = 'blob_key'
    self.mock.has_access.return_value = True
    self.mock.current_date_time.return_value = '2021-01-01 00:00:00 UTC'
    self.mock.utcnow.return_value = datetime.datetime(2021, 1, 1)
    self.mock.get_user_email.return_value = 'uploader@email'

    data_types.FuzzTarget(
        engine='libFuzzer', project='test-project', binary='binary').put()

    data_types.Job(
        name='libfuzzer_proj_external',
        environment_string='PROJECT_NAME = proj',
        platform='LINUX',
        external_reproduction_topic='topic',
        external_updates_subscription='sub').put()

    self.app = flask.Flask('testflask')
    self.app.add_url_rule(
        '/', view_func=upload_testcase.UploadHandlerOAuth.as_view(''))

  def _read_test_data(self, name):
    """Helper function to read test data."""
    with open(os.path.join(DATA_DIRECTORY, name), 'r') as handle:
      return handle.read()

  def assert_dict_has_items(self, expected, actual):
    """Assert that all items in `expected` are in `actual`."""
    for key, value in expected.items():
      self.assertEqual(value, actual[key], msg=f'For attribute {key}')

  def test_external_upload_oom(self):
    """Test external upload (oom)."""
    stacktrace = self._read_test_data('oom.txt')
    with self.app.test_client() as client:
      response = client.post(
          '/',
          data={
              'job': 'libfuzzer_proj_external',
              'target': 'target',
              'stacktrace': stacktrace,
              'revision': 1337,
              'file': (io.BytesIO(b'contents'), 'file'),
          })

    self.assertDictEqual({
        'id': '2',
        'uploadUrl': 'http://localhost//upload-testcase/upload-oauth'
    }, response.json)

    testcase = data_handler.get_testcase_by_id(2)
    self.assert_dict_has_items({
        'absolute_path': 'input',
        'additional_metadata': '{"fuzzer_binary_name": "target", '
                               '"uploaded_additional_args": "%TESTCASE%"}',
        'archive_filename': None,
        'archive_state': 0,
        'binary_flag': False,
        'bug_information': '',
        'comments': '[2021-01-01 00:00:00 UTC] uploader@email: '
                    'External testcase upload.\n',
        'crash_address': '',
        'crash_revision': 1337,
        'crash_stacktrace': stacktrace,
        'crash_state': 'target\n',
        'crash_type': 'Out-of-memory',
        'disable_ubsan': False,
        'duplicate_of': None,
        'fixed': '',
        'flaky_stack': False,
        'fuzzed_keys': 'blob_key',
        'fuzzer_name': 'libFuzzer',
        'gestures': [],
        'group_bug_information': 0,
        'group_id': 0,
        'has_bug_flag': False,
        'http_flag': False,
        'impact_beta_version': None,
        'impact_beta_version_likely': None,
        'impact_stable_version': None,
        'impact_stable_version_likely': None,
        'is_a_duplicate_flag': False,
        'is_impact_set_flag': None,
        'is_leader': False,
        'job_type': 'libfuzzer_proj_external',
        'last_tested_crash_stacktrace': None,
        'minidump_keys': None,
        'minimized_arguments': '',
        'minimized_keys': 'NA',
        'one_time_crasher_flag': False,
        'open': True,
        'overridden_fuzzer_name': 'libFuzzer_proj_target',
        'platform': 'linux',
        'platform_id': 'linux',
        'project_name': 'proj',
        'queue': None,
        'redzone': 128,
        'regression': 'NA',
        'security_flag': False,
        'security_severity': None,
        'status': 'Processed',
        'symbolized': False,
        'timeout_multiplier': 1.0,
        'timestamp': datetime.datetime(2021, 1, 1),
        'triaged': False,
        'uploader_email': 'uploader@email',
        'window_argument': ''
    }, testcase._to_dict())

    metadata = data_types.TestcaseUploadMetadata.query(
        data_types.TestcaseUploadMetadata.testcase_id ==
        testcase.key.id()).get()
    self.assertIsNotNone(metadata)
    self.assertDictEqual({
        'additional_metadata_string': None,
        'blobstore_key': 'blob_key',
        'bot_name': None,
        'bug_information': '',
        'bug_summary_update_flag': False,
        'bundled': False,
        'duplicate_of': None,
        'filename': 'input',
        'original_blobstore_key': 'blob_key',
        'path_in_archive': None,
        'quiet_flag': False,
        'retries': None,
        'security_flag': False,
        'status': 'Confirmed',
        'testcase_id': 2,
        'timeout': 0,
        'timestamp': datetime.datetime(2021, 1, 1, 0, 0),
        'uploader_email': 'uploader@email'
    }, metadata._to_dict())

  def test_external_upload_uaf(self):
    """Test external upload (uaf)."""
    stacktrace = self._read_test_data('uaf.txt')
    with self.app.test_client() as client:
      response = client.post(
          '/',
          data={
              'job': 'libfuzzer_proj_external',
              'target': 'target',
              'stacktrace': stacktrace,
              'revision': 1337,
              'file': (io.BytesIO(b'contents'), 'file'),
          })

    self.assertDictEqual({
        'id': '2',
        'uploadUrl': 'http://localhost//upload-testcase/upload-oauth'
    }, response.json)

    testcase = data_handler.get_testcase_by_id(2)
    self.assert_dict_has_items({
        'absolute_path':
            'input',
        'additional_metadata':
            '{"fuzzer_binary_name": "target", '
            '"uploaded_additional_args": "%TESTCASE%"}',
        'archive_filename':
            None,
        'archive_state':
            0,
        'binary_flag':
            False,
        'bug_information':
            '',
        'comments':
            '[2021-01-01 00:00:00 UTC] uploader@email: '
            'External testcase upload.\n',
        'crash_address':
            '0x60f00003b280',
        'crash_revision':
            1337,
        'crash_stacktrace':
            stacktrace,
        'crash_state': ('blink::InputTypeView::element\n'
                        'blink::TextFieldInputType::didSetValueByUserEdit\n'
                        'blink::TextFieldInputType::subtreeHasChanged\n'),
        'crash_type':
            'Heap-use-after-free\nREAD 8',
        'disable_ubsan':
            False,
        'duplicate_of':
            None,
        'fixed':
            '',
        'flaky_stack':
            False,
        'fuzzed_keys':
            'blob_key',
        'fuzzer_name':
            'libFuzzer',
        'gestures': [],
        'group_bug_information':
            0,
        'group_id':
            0,
        'has_bug_flag':
            False,
        'http_flag':
            False,
        'impact_beta_version':
            None,
        'impact_beta_version_likely':
            None,
        'impact_stable_version':
            None,
        'impact_stable_version_likely':
            None,
        'is_a_duplicate_flag':
            False,
        'is_impact_set_flag':
            None,
        'is_leader':
            False,
        'job_type':
            'libfuzzer_proj_external',
        'last_tested_crash_stacktrace':
            None,
        'minidump_keys':
            None,
        'minimized_arguments':
            '',
        'minimized_keys':
            'NA',
        'one_time_crasher_flag':
            False,
        'open':
            True,
        'overridden_fuzzer_name':
            'libFuzzer_proj_target',
        'platform':
            'linux',
        'platform_id':
            'linux',
        'project_name':
            'proj',
        'queue':
            None,
        'redzone':
            128,
        'regression':
            'NA',
        'security_flag':
            True,
        'security_severity':
            None,
        'status':
            'Processed',
        'symbolized':
            False,
        'timeout_multiplier':
            1.0,
        'timestamp':
            datetime.datetime(2021, 1, 1),
        'triaged':
            False,
        'uploader_email':
            'uploader@email',
        'window_argument':
            ''
    }, testcase._to_dict())

    metadata = data_types.TestcaseUploadMetadata.query(
        data_types.TestcaseUploadMetadata.testcase_id ==
        testcase.key.id()).get()
    self.assertIsNotNone(metadata)
    self.assertDictEqual({
        'additional_metadata_string': None,
        'blobstore_key': 'blob_key',
        'bot_name': None,
        'bug_information': '',
        'bug_summary_update_flag': False,
        'bundled': False,
        'duplicate_of': None,
        'filename': 'input',
        'original_blobstore_key': 'blob_key',
        'path_in_archive': None,
        'quiet_flag': False,
        'retries': None,
        'security_flag': True,
        'status': 'Confirmed',
        'testcase_id': 2,
        'timeout': 0,
        'timestamp': datetime.datetime(2021, 1, 1, 0, 0),
        'uploader_email': 'uploader@email'
    }, metadata._to_dict())

  def test_external_duplicate(self):
    """Test uploading a duplicate."""
    existing = data_types.Testcase(
        crash_address='',
        crash_state='target\n',
        crash_type='Out-of-memory',
        project_name='proj',
        minimized_keys='NA',
        security_flag=False)
    existing.put()

    stacktrace = self._read_test_data('oom.txt')
    with self.app.test_client() as client:
      response = client.post(
          '/',
          data={
              'job': 'libfuzzer_proj_external',
              'target': 'target',
              'stacktrace': stacktrace,
              'revision': 1337,
              'file': (io.BytesIO(b'contents'), 'file'),
          })

    self.assertDictEqual({
        'id': '3',
        'uploadUrl': 'http://localhost//upload-testcase/upload-oauth'
    }, response.json)

    testcase = data_handler.get_testcase_by_id(3)
    self.assert_dict_has_items({
        'absolute_path': 'input',
        'additional_metadata': '{"fuzzer_binary_name": "target", '
                               '"uploaded_additional_args": "%TESTCASE%"}',
        'archive_filename': None,
        'archive_state': 0,
        'binary_flag': False,
        'bug_information': '',
        'comments': '[2021-01-01 00:00:00 UTC] uploader@email: '
                    'External testcase upload.\n',
        'crash_address': '',
        'crash_revision': 1337,
        'crash_stacktrace': stacktrace,
        'crash_state': 'target\n',
        'crash_type': 'Out-of-memory',
        'disable_ubsan': False,
        'duplicate_of': 2,
        'fixed': 'NA',
        'flaky_stack': False,
        'fuzzed_keys': 'blob_key',
        'fuzzer_name': 'libFuzzer',
        'gestures': [],
        'group_bug_information': 0,
        'group_id': 0,
        'has_bug_flag': False,
        'http_flag': False,
        'impact_beta_version': None,
        'impact_beta_version_likely': False,
        'impact_stable_version': None,
        'impact_stable_version_likely': False,
        'is_a_duplicate_flag': True,
        'is_impact_set_flag': False,
        'is_leader': False,
        'job_type': 'libfuzzer_proj_external',
        'last_tested_crash_stacktrace': None,
        'minidump_keys': None,
        'minimized_arguments': '',
        'minimized_keys': 'NA',
        'one_time_crasher_flag': False,
        'open': False,
        'overridden_fuzzer_name': 'libFuzzer_proj_target',
        'platform': 'linux',
        'platform_id': 'linux',
        'project_name': 'proj',
        'queue': None,
        'redzone': 128,
        'regression': 'NA',
        'security_flag': False,
        'security_severity': None,
        'status': 'Duplicate',
        'symbolized': False,
        'timeout_multiplier': 1.0,
        'timestamp': datetime.datetime(2021, 1, 1),
        'triaged': True,
        'uploader_email': 'uploader@email',
        'window_argument': ''
    }, testcase._to_dict())

    metadata = data_types.TestcaseUploadMetadata.query(
        data_types.TestcaseUploadMetadata.testcase_id ==
        testcase.key.id()).get()
    self.assertIsNotNone(metadata)
    self.assertDictEqual({
        'additional_metadata_string': None,
        'blobstore_key': 'blob_key',
        'bot_name': None,
        'bug_information': '',
        'bug_summary_update_flag': False,
        'bundled': False,
        'duplicate_of': 2,
        'filename': 'input',
        'original_blobstore_key': 'blob_key',
        'path_in_archive': None,
        'quiet_flag': False,
        'retries': None,
        'security_flag': False,
        'status': 'Duplicate',
        'testcase_id': 3,
        'timeout': 0,
        'timestamp': datetime.datetime(2021, 1, 1, 0, 0),
        'uploader_email': 'uploader@email'
    }, metadata._to_dict())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzing strategy selection cron job.

Runs multi-armed bandit experiments for fuzzing strategy selection.
In particular, this is a Boltzman Exploration (softmax) implementation
of multi-armed bandit experiments. Queries from bigquery to update
multi-armed bandit probability values based on the new edges for various
combined strategies. In the upload_bandit_weights function, we can change
metric to be for edges, crash, features, or units. Currently based on new
edges."""

from collections import namedtuple

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.fuzzing import strategy
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler

# After experimentation with high, low, and medium temperature parameters, we
# decided on .15.
TEMPERATURE_PARAMETER = .15

# Maintain a list of strategies to include in query for each fuzzing engine.
# Keep this strategy order for strategy combination tracking as strategy
# combinations are tracked as strings.
libfuzzer_query_strategy_list = [
    strategy_tuple for strategy_tuple in strategy.LIBFUZZER_STRATEGY_LIST
    if not strategy_tuple.manually_enable
]

afl_query_strategy_list = [
    strategy_tuple for strategy_tuple in strategy.AFL_STRATEGY_LIST
    if not strategy_tuple.manually_enable
]

# A tuple of engine name and corresponding strategies to include in multi-armed
# bandit query.
Engine = namedtuple('Engine', 'name query_strategy_list performance_metric')

LIBFUZZER_ENGINE = Engine(
    name='libFuzzer',
    query_strategy_list=libfuzzer_query_strategy_list,
    performance_metric='new_edges')

AFL_ENGINE = Engine(
    name='afl',
    query_strategy_list=afl_query_strategy_list,
    performance_metric='new_units_generated')

ENGINE_LIST = [LIBFUZZER_ENGINE, AFL_ENGINE]

# BigQuery query for calculating multi-armed bandit probabilities for
# various strategies using a Boltzman Exploration (softmax) model.

# Averages standardized new_edges feature over each strategy for expected
# new_edges metric for each strategy.
# See https://www.cs.mcgill.ca/~vkules/bandits.pdf for formula.

# TODO(mukundv): Change query once we decide on a temperature parameter and
# final implementation.

BANDIT_PROBABILITY_QUERY_FORMAT = """
(SELECT
    /* Calculate bandit weights from calculated exponential values. */
    strategy,
    strategy_exp / exp_sum AS bandit_weight
  FROM (
    SELECT
      EXP(strategy_avg_{performance_metric} / temperature) AS strategy_exp,
      SUM(EXP(strategy_avg_{performance_metric} / temperature)) OVER() AS exp_sum,
      strategy
    FROM (
      SELECT
        /* Standardize the new edges data and take averages per strategy. */
        AVG(({performance_metric} - overall_avg_{performance_metric}) / overall_stddev_{performance_metric}) AS strategy_avg_{performance_metric},
        strategy,
        /* Change temperature parameter here. */
        {temperature_value} AS temperature
      FROM (
        SELECT
          fuzzer,
          CONCAT({strategies}) AS strategy,
          fuzzer_stddev,
          AVG({performance_metric}) OVER() AS overall_avg_{performance_metric},
          STDDEV({performance_metric}) OVER() AS overall_stddev_{performance_metric},
          {performance_metric},
          strategy_selection_method
        FROM (
          SELECT
            fuzzer,
            {strategies_subquery}
            STDDEV({performance_metric}) OVER(PARTITION BY fuzzer) AS fuzzer_stddev,
            {performance_metric},
            strategy_selection_method
          FROM
            {engine}_stats.TestcaseRun
          WHERE
            /* Query results from the past 5 days. Change as needed. */
            DATE_DIFF(CAST(CURRENT_TIMESTAMP() AS DATE), CAST(_PARTITIONTIME AS DATE), DAY) < 6 )
        WHERE
          /* Filter for unstable targets. */
          fuzzer_stddev < 50)
      GROUP BY
        strategy)))
"""

STRATEGY_SUBQUERY_FORMAT = """
IF
  (strategy_{strategy_name} > 0,
    "{strategy_name},",
    "") AS strategy_{strategy_name},
"""


def _query_multi_armed_bandit_probabilities(engine):
  """Get query results.

  Queries above BANDIT_PROBABILITY_QUERY and yields results
  from bigquery. This query is sorted by strategies implemented."""
  strategy_names_list = [
      strategy_entry.name for strategy_entry in engine.query_strategy_list
  ]
  strategies_subquery = '\n'.join([
      STRATEGY_SUBQUERY_FORMAT.format(strategy_name=strategy_name)
      for strategy_name in strategy_names_list
  ])
  client = big_query.Client()
  strategies = ','.join(
      ['strategy_' + strategy_name for strategy_name in strategy_names_list])
  formatted_query = BANDIT_PROBABILITY_QUERY_FORMAT.format(
      performance_metric=engine.performance_metric,
      temperature_value=TEMPERATURE_PARAMETER,
      strategies=strategies,
      strategies_subquery=strategies_subquery,
      engine=engine.name)
  return client.query(query=formatted_query).rows


def _store_probabilities_in_bigquery(engine, data):
  """Update a bigquery table containing the daily updated
  probability distribution over strategies."""
  bigquery_data = []

  # TODO(mukundv): Update once we choose a temperature parameter for final
  # implementation.
  for row in data:
    bigquery_row = {
        'strategy_name': row['strategy'],
        'probability': row['bandit_weight'],
        'engine': engine.name
    }
    bigquery_data.append(big_query.Insert(row=bigquery_row, insert_id=None))

  if bigquery_data:
    client = big_query.Client(
        dataset_id='main', table_id='fuzz_strategy_probability')
    client.insert(bigquery_data)
  else:
    logs.log('No fuzz strategy distribution data was found to upload to '
             'BigQuery.')


def _query_and_upload_strategy_probabilities(engine):
  """Uploads queried data into datastore.

  Calls query functions and uploads query results
  to datastore to use as new probabilities. Probabilities
  are based on new_edges feature."""
  strategy_data = []
  data = _query_multi_armed_bandit_probabilities(engine)
  logs.log('Queried distribution for {}.'.format(engine.name))

  # TODO(mukundv): Update once we choose a temperature parameter for final
  # implementation.
  for row in data:
    curr_strategy = data_types.FuzzStrategyProbability()
    curr_strategy.strategy_name = str(row['strategy'])
    curr_strategy.probability = float(row['bandit_weight'])
    curr_strategy.engine = engine.name
    strategy_data.append(curr_strategy)

  query = data_types.FuzzStrategyProbability.query(
      data_types.FuzzStrategyProbability.engine == engine.name)
  ndb_utils.delete_multi(
      [entity.key for entity in ndb_utils.get_all_from_query(query)])
  ndb_utils.put_multi(strategy_data)
  logs.log('Uploaded queried distribution to ndb for {}'.format(engine.name))
  _store_probabilities_in_bigquery(engine, data)
  logs.log('Uploaded queried distribution to BigQuery for {}'.format(
      engine.name))


class Handler(base_handler.Handler):
  """Cron job handler for fuzz strategy selection.

  Handler to periodically update fuzz strategy bandit probabilities
  based on a performance metric (currently based on new_edges)."""

  @handler.cron()
  def get(self):
    """Process all fuzz targets and update FuzzStrategy weights."""
    for engine in ENGINE_LIST:
      _query_and_upload_strategy_probabilities(engine)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common functionality for Crash Stats cron job, backend, and frontend."""

import datetime
import json
import math

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.system import environment

SQL = """
WITH
  # Deduplicate rows in case build_crash_stats runs twice on the same hour.
  uniqueRows AS (
    SELECT
      crash_type, crash_state, security_flag, hour, parent_fuzzer_name,
      fuzzer_name, job_type, revision, parent_platform, platform, project,
      MAX(count) AS count,
      MAX(min_crash_time_in_ms) AS min_crash_time_in_ms,
      MAX(max_crash_time_in_ms) AS max_crash_time_in_ms,
      MAX(sum_crash_time_in_ms) AS sum_crash_time_in_ms,
      MAX(sum_square_crash_time_in_ms) AS sum_square_crash_time_in_ms,
      LOGICAL_OR(new_flag) AS new_flag,
      LOGICAL_OR(reproducible_flag) AS reproducible_flag
    FROM main.crash_stats
    WHERE {where_clause}
    GROUP BY
      crash_type, crash_state, security_flag, hour, parent_fuzzer_name,
      fuzzer_name, job_type, revision, parent_platform, platform, project
  ),
  groupByFieldAndIndex AS (
    SELECT
      project, crash_type, crash_state, security_flag, {group_by},
      CAST(FLOOR((hour - {remainder}) / {time_span}) AS INT64) as index,
      SUM(count) AS count,
      MIN(min_crash_time_in_ms) AS min_crash_time_in_ms,
      MAX(max_crash_time_in_ms) AS max_crash_time_in_ms,
      SUM(sum_crash_time_in_ms) AS sum_crash_time_in_ms,
      SUM(sum_square_crash_time_in_ms) AS sum_square_crash_time_in_ms,
      LOGICAL_OR(new_flag) AS is_new,
      LOGICAL_OR(reproducible_flag) AS is_reproducible
    FROM uniqueRows
    GROUP BY
      project, crash_type, crash_state, security_flag, {group_by}, index
    ORDER BY {group_by} ASC, index DESC
  ),
  groupByIndex AS (
    SELECT
      SUM(count) AS total_count, index
    FROM groupByFieldAndIndex
    GROUP BY index
  ),
  withTotal AS (
    SELECT *
    FROM groupByFieldAndIndex LEFT OUTER JOIN groupByIndex
    USING (index)
  ),
  groupByField AS (
    SELECT
      project, crash_type, crash_state, security_flag,
      CAST({group_by} AS STRING) AS {group_by},
      SUM(count) AS total_count,
      MIN(min_crash_time_in_ms) AS min_crash_time_in_ms,
      MAX(max_crash_time_in_ms) AS max_crash_time_in_ms,
      SUM(sum_crash_time_in_ms) AS sum_crash_time_in_ms,
      SUM(sum_square_crash_time_in_ms) AS sum_square_crash_time_in_ms,
      LOGICAL_OR(is_new) AS is_new,
      LOGICAL_OR(is_reproducible) AS is_reproducible,
      MIN(index) AS first_index,
      ARRAY_AGG(STRUCT(
        index,
        is_new,
        CAST(((count / withTotal.total_count) * 100) AS INT64) AS percent,
        count AS count
      )) AS indices
    FROM withTotal
    GROUP BY
      project, crash_type, crash_state, security_flag, {group_by}
    ORDER BY {sort_by} DESC, total_count DESC
  )

SELECT
  project, crash_type, crash_state, security_flag,
  SUM(total_count) AS total_count,
  MIN(min_crash_time_in_ms) AS min_crash_time_in_ms,
  MAX(max_crash_time_in_ms) AS max_crash_time_in_ms,
  SUM(sum_crash_time_in_ms) AS sum_crash_time_in_ms,
  SUM(sum_square_crash_time_in_ms) AS sum_square_crash_time_in_ms,
  LOGICAL_OR(is_new) AS is_new,
  LOGICAL_OR(is_reproducible) AS is_reproducible,
  MIN(first_index) AS first_index,
  ARRAY_AGG(STRUCT({group_by} AS name, indices)) AS `groups`
FROM groupByField
GROUP BY
  project, crash_type, crash_state, security_flag
{group_having_clause}
ORDER BY {sort_by} DESC, total_count DESC
"""


def get_remainder_for_index(true_end, time_span):
  """Get remainder. This should be tested together with
    convert_index_to_hour."""
  # The remainder needs +1 because the cut-off is at the end of true_end.
  # For example, if the true_end is 49, the 1st day is 2 to 25 and the
  # 2nd day is 26, to 49.
  return (true_end % time_span) + 1


def convert_index_to_hour(index, time_span, remainder):
  """Convert index to hour."""
  # This needs -1 because the end hour is inclusive. For example, if the period
  # represents [2, 26), the end hour is 25.
  #
  # Index is added 1 because, in our SQL, we subtract the remainder, divide,
  # and floor. So, in order to get the original hour, we need to add 1 to the
  # index.
  return ((index + 1) * time_span) + remainder - 1


def get(end, days, block, group_by, where_clause, group_having_clause, sort_by,
        offset, limit):
  """Query from BigQuery given the params."""
  if where_clause:
    where_clause = '(%s) AND ' % where_clause

  start = end - (days * 24) + 1

  where_clause += '(hour BETWEEN %d AND %d) AND ' % (start, end)
  where_clause += ('(_PARTITIONTIME BETWEEN TIMESTAMP_TRUNC("%s", DAY) '
                   'AND TIMESTAMP_TRUNC("%s", DAY))' %
                   (get_datetime(start).strftime('%Y-%m-%d'),
                    get_datetime(end).strftime('%Y-%m-%d')))

  time_span = 1 if block == 'hour' else 24
  remainder = get_remainder_for_index(end, time_span)

  if group_having_clause:
    group_having_clause = 'HAVING ' + group_having_clause

  if (not big_query.VALID_FIELD_NAME_REGEX.match(group_by) or
      not big_query.VALID_FIELD_NAME_REGEX.match(sort_by)):
    raise ValueError('Invalid group_by or sort_by')

  sql = SQL.format(
      time_span=time_span,
      remainder=remainder,
      group_by=group_by,
      where_clause=where_clause,
      group_having_clause=group_having_clause,
      sort_by=sort_by)

  client = big_query.Client()
  result = client.query(query=sql, offset=offset, limit=limit)

  items = []
  for row in result.rows:
    avg_crash_time_in_ms = row['sum_crash_time_in_ms'] // row['total_count']

    for group in row['groups']:
      for index in group['indices']:
        index['hour'] = convert_index_to_hour(index['index'], time_span,
                                              remainder)

    items.append({
        'projectName': row['project'],
        'crashType': row['crash_type'],
        'crashState': row['crash_state'],
        'isSecurity': row['security_flag'],
        'isReproducible': row['is_reproducible'],
        'isNew': row['is_new'],
        'totalCount': row['total_count'],
        'crashTime': {
            'min':
                row['min_crash_time_in_ms'],
            'max':
                row['max_crash_time_in_ms'],
            'avg':
                avg_crash_time_in_ms,
            'std':
                math.sqrt(
                    (row['sum_square_crash_time_in_ms'] // row['total_count']) -
                    (avg_crash_time_in_ms * avg_crash_time_in_ms))
        },
        'groups': row['groups'],
        'days': days,
        'block': block,
        'end': end + 1  # Convert to UI's end.
    })
  return result.total_count, items


def get_datetime(hours):
  """Get datetime obj from hours from epoch."""
  return datetime.datetime.utcfromtimestamp(hours * 60 * 60)


def _get_first_or_last_successful_hour(is_last):
  """Get the first successful hour."""
  order = data_types.BuildCrashStatsJobHistory.end_time_in_hours
  if is_last:
    order = -order

  item = data_types.BuildCrashStatsJobHistory.query().order(order).get()
  if not item:
    return None

  return item.end_time_in_hours


def get_last_successful_hour():
  """Get the last hour that ran successfully. We want to run the next hour."""
  return _get_first_or_last_successful_hour(is_last=True)


def get_min_hour():
  """Get the first hour that ran successfully (for the date-time picker)."""
  hour = _get_first_or_last_successful_hour(is_last=False)

  # `hour` is None when we haven't run build_crash_stats at all.
  # Therefore, there's no crash stats data.
  #
  # On the UI, the date-time picker choose a point of time. Therefore,
  # if we choose, say, 3pm, this means we want the crash stats until 2:59pm.
  # Therefore, we need to increment by 1.
  return (hour or 0) + 1


def get_max_hour():
  """Get the last hour that can be selected by the date-time picker."""
  hour = get_last_successful_hour()

  # `hour` is None when we haven't run build_crash_stats at all.
  # Therefore, there's no crash stats data.
  #
  # On the UI, the date-time picker choose a point of time. Therefore,
  # if we choose, say, 3pm, this means we want the crash stats until 2:59pm.
  # Therefore, we need to increment by 1.
  return (hour or 0) + 1


@environment.local_noop
def get_last_crash_time(testcase):
  """Return timestamp for last crash with same crash params as testcase."""
  client = big_query.Client()

  where_clause = ('crash_type = {crash_type} AND '
                  'crash_state = {crash_state} AND '
                  'security_flag = {security_flag} AND '
                  'project = {project}').format(
                      crash_type=json.dumps(testcase.crash_type),
                      crash_state=json.dumps(testcase.crash_state),
                      security_flag=json.dumps(testcase.security_flag),
                      project=json.dumps(testcase.project_name),
                  )

  sql = """
SELECT hour
FROM main.crash_stats
WHERE {where_clause}
ORDER by hour DESC
LIMIT 1
""".format(where_clause=where_clause)

  result = client.query(query=sql)
  if result and result.rows:
    return get_datetime(result.rows[0]['hour'])

  return None



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Variant task for analyzing testcase variants with a different job."""

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.crash_analysis.crash_comparer import CrashComparer
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment


def _get_variant_testcase_for_job(testcase, job_type):
  """Return a testcase entity for variant task use. This changes the fuzz
  target params for a particular fuzzing engine."""
  if testcase.job_type == job_type:
    # Update stack operation on same testcase.
    return testcase

  if not environment.is_engine_fuzzer_job(testcase.job_type):
    # For blackbox fuzzer testcases, there is no change of fuzzer required.
    return testcase

  engine_name = environment.get_engine_for_job(job_type)
  project = data_handler.get_project_name(job_type)
  binary_name = testcase.get_metadata('fuzzer_binary_name')
  fully_qualified_fuzzer_name = data_types.fuzz_target_fully_qualified_name(
      engine_name, project, binary_name)

  variant_testcase = data_types.clone_entity(testcase)
  variant_testcase.key = testcase.key
  variant_testcase.fuzzer_name = engine_name
  variant_testcase.overridden_fuzzer_name = fully_qualified_fuzzer_name
  variant_testcase.job_type = job_type

  # Remove put() method to avoid updates. DO NOT REMOVE THIS.
  variant_testcase.put = lambda: None

  return variant_testcase


def execute_task(testcase_id, job_type):
  """Run a test case with a different job type to see if they reproduce."""
  testcase = data_handler.get_testcase_by_id(testcase_id)
  if not testcase:
    return

  if (environment.is_engine_fuzzer_job(testcase.job_type) !=
      environment.is_engine_fuzzer_job(job_type)):
    # We should never reach here. But in case we do, we should bail out as
    # otherwise we will run into exceptions.
    return

  # Use a cloned testcase entity with different fuzz target paramaters for
  # a different fuzzing engine.
  original_job_type = testcase.job_type
  testcase = _get_variant_testcase_for_job(testcase, job_type)

  # Setup testcase and its dependencies.
  file_list, _, testcase_file_path = setup.setup_testcase(testcase, job_type)
  if not file_list:
    return

  # Set up a custom or regular build. We explicitly omit the crash revision
  # since we want to test against the latest build here.
  build_manager.setup_build()

  # Check if we have an application path. If not, our build failed to setup
  # correctly.
  if not build_manager.check_app_path():
    testcase = data_handler.get_testcase_by_id(testcase_id)
    data_handler.update_testcase_comment(
        testcase, data_types.TaskState.ERROR,
        'Build setup failed with job: ' + job_type)
    return

  # Disable gestures if we're running on a different platform from that of
  # the original test case.
  use_gestures = testcase.platform == environment.platform().lower()

  # Reproduce the crash.
  app_path = environment.get_value('APP_PATH')
  command = testcase_manager.get_command_line_for_application(
      testcase_file_path, app_path=app_path, needs_http=testcase.http_flag)
  test_timeout = environment.get_value('TEST_TIMEOUT', 10)
  revision = environment.get_value('APP_REVISION')
  result = testcase_manager.test_for_crash_with_retries(
      testcase,
      testcase_file_path,
      test_timeout,
      http_flag=testcase.http_flag,
      use_gestures=use_gestures,
      compare_crash=False)

  if result.is_crash() and not result.should_ignore():
    crash_state = result.get_state()
    crash_type = result.get_type()
    security_flag = result.is_security_issue()

    gestures = testcase.gestures if use_gestures else None
    one_time_crasher_flag = not testcase_manager.test_for_reproducibility(
        testcase.fuzzer_name, testcase.actual_fuzzer_name(), testcase_file_path,
        crash_state, security_flag, test_timeout, testcase.http_flag, gestures)
    if one_time_crasher_flag:
      status = data_types.TestcaseVariantStatus.FLAKY
    else:
      status = data_types.TestcaseVariantStatus.REPRODUCIBLE

    crash_comparer = CrashComparer(crash_state, testcase.crash_state)
    is_similar = (
        crash_comparer.is_similar() and security_flag == testcase.security_flag)

    unsymbolized_crash_stacktrace = result.get_stacktrace(symbolized=False)
    symbolized_crash_stacktrace = result.get_stacktrace(symbolized=True)
    crash_stacktrace_output = utils.get_crash_stacktrace_output(
        command, symbolized_crash_stacktrace, unsymbolized_crash_stacktrace)
  else:
    status = data_types.TestcaseVariantStatus.UNREPRODUCIBLE
    is_similar = False
    crash_type = None
    crash_state = None
    security_flag = False
    crash_stacktrace_output = 'No crash occurred.'

  if original_job_type == job_type:
    # This case happens when someone clicks 'Update last tested stacktrace using
    # trunk build' button.
    testcase = data_handler.get_testcase_by_id(testcase_id)
    testcase.last_tested_crash_stacktrace = (
        data_handler.filter_stacktrace(crash_stacktrace_output))
    testcase.set_metadata(
        'last_tested_crash_revision', revision, update_testcase=True)
  else:
    # Regular case of variant analysis.
    variant = data_handler.get_testcase_variant(testcase_id, job_type)
    variant.status = status
    variant.revision = revision
    variant.crash_type = crash_type
    variant.crash_state = crash_state
    variant.security_flag = security_flag
    variant.is_similar = is_similar
    variant.platform = environment.platform().lower()
    # Explicitly skipping crash stacktrace for now as it make entities larger
    # and we plan to use only crash paramaters in UI.
    variant.put()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""oss_fuzz_apply_ccs tests."""
import unittest

import flask
from google.cloud import ndb
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import oss_fuzz_generate_certs


@test_utils.with_cloud_emulators('datastore')
class OssFuzzGenerateCertsTest(unittest.TestCase):
  """Test oss_fuzz_generate_certs."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'handlers.base_handler.Handler.is_cron',
    ])

    self.mock.is_cron.return_value = True
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/generate-certs',
        view_func=oss_fuzz_generate_certs.Handler.as_view('/generate-certs'))
    self.app = webtest.TestApp(flaskapp)

    data_types.OssFuzzProject(name='project1').put()
    data_types.OssFuzzProject(name='project2').put()

    data_types.WorkerTlsCert(
        id='project2',
        project_name='project2',
        cert_contents=b'cert_contents',
        key_contents=b'key_contents').put()

  def test_execute(self):
    """Tests executing of cron job."""
    # Defer import to avoid issues on Python 2.
    from OpenSSL import crypto

    self.app.get('/generate-certs')

    # New cert.
    tls_cert = ndb.Key(data_types.WorkerTlsCert, 'project1').get()
    cert = crypto.load_certificate(crypto.FILETYPE_PEM, tls_cert.cert_contents)
    self.assertEqual('US', cert.get_subject().C)
    self.assertEqual('*.c.test-clusterfuzz.internal', cert.get_subject().CN)
    self.assertEqual('project1', cert.get_subject().O)
    self.assertEqual(9001, cert.get_serial_number())
    self.assertEqual(b'20000101000000Z', cert.get_notBefore())
    self.assertEqual(b'21000101000000Z', cert.get_notAfter())

    private_key = crypto.load_privatekey(crypto.FILETYPE_PEM,
                                         tls_cert.key_contents)
    self.assertTrue(private_key.check())

    # Should be unchanged.
    tls_cert = ndb.Key(data_types.WorkerTlsCert, 'project2').get()
    self.assertEqual(b'cert_contents', tls_cert.cert_contents)
    self.assertEqual(b'key_contents', tls_cert.key_contents)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the issue_tracker_utils module."""

import unittest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from libs.issue_management import issue_tracker
from libs.issue_management import issue_tracker_utils


class IssueTrackerUtilsUrlTest(unittest.TestCase):
  """Issue tracker utils tests for URL handling methods."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.issue_management.issue_tracker_utils.'
        'get_issue_tracker_for_testcase',
        'libs.issue_management.issue_tracker.IssueTracker.issue_url',
    ])

  def test_get_issue_url(self):
    """Basic test for a case when testcase is associated with a bug."""
    testcase = data_types.Testcase()
    testcase.bug_information = '1337'

    test_issue_tracker = issue_tracker.IssueTracker()
    self.mock.get_issue_tracker_for_testcase.return_value = test_issue_tracker

    issue_tracker_utils.get_issue_url(testcase)
    self.mock.issue_url.assert_called_with(test_issue_tracker, '1337')

  def test_get_issue_url_group_bug(self):
    """Test for a case when testcase is associated with a group bug."""
    testcase = data_types.Testcase()
    testcase.group_bug_information = 31337

    test_issue_tracker = issue_tracker.IssueTracker()
    self.mock.get_issue_tracker_for_testcase.return_value = test_issue_tracker

    issue_tracker_utils.get_issue_url(testcase)
    self.mock.issue_url.assert_called_with(test_issue_tracker, '31337')

  def test_get_issue_url_no_bug(self):
    """Test for a case when testcase has no bugs associated with it."""
    testcase = data_types.Testcase()

    test_issue_tracker = issue_tracker.IssueTracker()
    self.mock.get_issue_tracker_for_testcase.return_value = test_issue_tracker

    issue_tracker_utils.get_issue_url(testcase)
    self.assertEqual(0, self.mock.issue_url.call_count)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test the stats.py script for AFL-based fuzzers."""

import copy
import os
import unittest

import mock
import six

from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers.afl import launcher
from clusterfuzz._internal.bot.fuzzers.afl import stats
from clusterfuzz._internal.fuzzing import strategy
from clusterfuzz._internal.tests.core.bot.fuzzers.afl.afl_launcher_test import \
    dont_use_strategies
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


def override_fail_retries(env_var, default_value=None):
  """Used to patch environment.getEnv to set FAIL_RETRIES."""
  return os.getenv(
      env_var, default=default_value) if env_var != 'FAIL_RETRIES' else 1


class StatsGetterTests(unittest.TestCase):
  """Tests for launcher.StatsGetter"""
  MANUAL_DICT_SIZE = 3
  NEW_UNITS_GENERATED = 2
  NEW_UNITS_ADDED = 1
  CORPUS_SIZE = 20
  ACTUAL_DURATION = 5

  @mock.patch('clusterfuzz._internal.system.environment.get_value',
              override_fail_retries)
  def setUp(self):

    def get_data_path(filename, is_in_output=True):
      """Returns absolute path of data files used in unittests."""
      input_or_output_dir = 'output' if is_in_output else 'input'
      return os.path.join(self.data_dir, input_or_output_dir, filename)

    self.data_dir = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), 'stats_data')
    self.fuzzer_stats_path = get_data_path('fuzzer_stats')
    self.fuzzer_stats_invalid_path = get_data_path('fuzzer_stats_invalid')
    self.dict_path = get_data_path('sample.dict', False)

    self.stats_getter = stats.StatsGetter(self.fuzzer_stats_path,
                                          self.dict_path)

    self.stats_getter.set_afl_stats()
    dont_use_strategies(self)
    test_helpers.patch(
        self,
        ['clusterfuzz._internal.bot.fuzzers.engine_common.is_lpm_fuzz_target'])
    self.mock.is_lpm_fuzz_target.return_value = True
    self.strategies = launcher.FuzzingStrategies(None)

    self.maxDiff = None  # pylint: disable=invalid-name

  def test_set_afl_stats(self):
    """Tests that set_afl_stats() parses a fuzzer_stats file properly."""
    expected_stats = {
        'afl_version': '2.38',
        'bitmap_cvg': '0.00',
        'cur_path': '1',
        'cycles_done': '1376',
        'execs_done': '621037',
        'execs_per_sec': '76.19',
        'execs_since_crash': '621037',
        'exec_timeout': '20',
        'fuzzer_pid': '37784',
        'last_crash': '0',
        'last_hang': '1504640708',
        'last_path': '1504640693',
        'last_update': '1504640710',
        'max_depth': '2',
        'paths_favored': '2',
        'paths_found': '1',
        'paths_imported': '0',
        'paths_total': '2',
        'pending_favs': '0',
        'pending_total': '0',
        'stability': '100.00',
        'start_time': '1504640693',
        'unique_crashes': '0',
        'unique_hangs': '1',
        'variable_paths': '0',
    }

    self.assertEqual(self.stats_getter.afl_stats, expected_stats)

  @mock.patch('clusterfuzz._internal.system.environment.get_value',
              override_fail_retries)
  def _set_stats(self):
    """Helper function that calls self.stats_getter.set_stats with default
    values and returns the result.
    """
    return self.stats_getter.set_stats(
        self.ACTUAL_DURATION,
        self.NEW_UNITS_GENERATED,
        self.NEW_UNITS_ADDED,
        self.CORPUS_SIZE,
        self.strategies,
        fuzzer_stderr='',
        afl_fuzz_output='')

  def test_dict_stats(self):
    """Tests that "dict_used" and "manual_dict_size" are set properly by
    set_stats() when self.stats_getter is given a valid dictionary.
    """
    actual_stats = self._set_stats()

    self.assertEqual(actual_stats['dict_used'], 1)
    self.assertEqual(actual_stats['manual_dict_size'], self.MANUAL_DICT_SIZE)

  def test_stats_without_dict(self):
    """Tests that "dict_used" and "manual_dict_size" are set properly by
    set_stats() when self.stats_getter is *not* given a dictionary.
    """
    self.stats_getter.dict_path = None
    actual_stats = self._set_stats()
    self.assertEqual(actual_stats['dict_used'], 0)
    self.assertEqual(actual_stats['manual_dict_size'], 0)

  def test_set_stats(self):
    """Tests that all stats are set properly by set_stats() when stats_getter
    is given a valid fuzzer_stats file.
    """
    actual_stats = self._set_stats()

    expected_stats = {
        'actual_duration': 5,
        'average_exec_per_sec': 124207,
        'bad_instrumentation': 0,
        'corpus_crash_count': 0,
        'corpus_size': 20,
        'crash_count': 0,
        'dict_used': 1,
        'log_lines_unwanted': 0,
        'manual_dict_size': 3,
        'new_units_added': 1,
        'new_units_generated': 2,
        'stability': 100.0,
        'startup_crash_count': 0,
        'strategy_selection_method': 'default',
        'timeout_count': 1,
        'timeout_limit': 20,
    }

    self.assertEqual(actual_stats, expected_stats)

  def test_set_stats_no_file(self):
    """Tests that set_stats() sets the correct stats or uses default stats when
    there is no fuzzer_stats file."""

    fuzzer_stats_path = self.fuzzer_stats_path + '.nonexistent'
    self.stats_getter = stats.StatsGetter(fuzzer_stats_path, self.dict_path)
    expected_stats = {
        'actual_duration': 5,
        'average_exec_per_sec': 0,
        'bad_instrumentation': 0,
        'corpus_crash_count': 0,
        'corpus_size': 20,
        'crash_count': 0,
        'dict_used': 1,
        'log_lines_unwanted': 0,
        'manual_dict_size': 3,
        'new_units_added': 1,
        'new_units_generated': 2,
        'stability': 0.0,
        'startup_crash_count': 0,
        'strategy_selection_method': 'default',
        'timeout_count': 0,
        'timeout_limit': 0,
    }
    self.assertEqual(self._set_stats(), expected_stats)

  def test_set_stats_invalid_file(self):
    """Tests that all stats are set properly by set_stats() when stats_getter
    is *not* given a valid fuzzer_stats file.
    """
    expected_stats = {
        'actual_duration': 5,
        'average_exec_per_sec': 0,
        'bad_instrumentation': 0,
        'corpus_crash_count': 0,
        'corpus_size': 20,
        'crash_count': 0,
        'dict_used': 1,
        'log_lines_unwanted': 0,
        'manual_dict_size': 3,
        'new_units_added': 1,
        'new_units_generated': 2,
        'stability': 0.0,
        'startup_crash_count': 0,
        'strategy_selection_method': 'default',
        'timeout_count': 0,
        'timeout_limit': 0,
    }

    self.stats_getter = stats.StatsGetter(self.fuzzer_stats_invalid_path,
                                          self.dict_path)

    actual_stats = self._set_stats()
    self.assertEqual(actual_stats, expected_stats)

  @mock.patch('clusterfuzz._internal.system.environment.get_value',
              override_fail_retries)
  def test_actual_duration_is_0(self):
    """Tests that average_exec_per_sec is set to 0 when actual_duration is 0."""
    self.stats_getter.set_stats(
        0,
        self.NEW_UNITS_GENERATED,
        self.NEW_UNITS_ADDED,
        self.CORPUS_SIZE,
        self.strategies,
        fuzzer_stderr='',
        afl_fuzz_output='')

    self.assertEqual(self.stats_getter.stats['average_exec_per_sec'], 0)

  @mock.patch('clusterfuzz._internal.system.environment.get_value',
              override_fail_retries)
  def test_log_lines_unwanted(self):
    """Tests that average_exec_per_sec is set to 0 when actual_duration is 0."""
    with open(os.path.join(self.data_dir, 'unwanted_logging.txt')) as f:
      fuzzer_stderr = f.read()

    self.stats_getter.set_stats(
        self.ACTUAL_DURATION,
        self.NEW_UNITS_GENERATED,
        self.NEW_UNITS_ADDED,
        self.CORPUS_SIZE,
        self.strategies,
        fuzzer_stderr,
        afl_fuzz_output='')

    self.assertEqual(self.stats_getter.stats['log_lines_unwanted'], 18)

  def test_get_afl_stat(self):
    """Tests that get_afl_stat works as intended when a stat that is in
    fuzzer_stats is asked for."""
    self.assertEqual(self.stats_getter.get_afl_stat('paths_total'), 2)

  def test_get_missing_afl_stat(self):
    """Tests that get_afl_stat returns 0 when a stat that is *not* in
      fuzzer_stats is asked for."""
    self.assertEqual(self.stats_getter.get_afl_stat('MISSING_STAT'), 0)

  def test_correct_types(self):
    """Tests that the types of the stats set by StatsGetter.set_stats are the
    same as the default ones."""
    default_stats = copy.copy(self.stats_getter.stats)
    actual_stats = self._set_stats()
    for stat_key, stat_value in six.iteritems(default_stats):
      self.assertEqual(type(stat_value), type(actual_stats[stat_key]))

  def test_set_strategy_stats(self):
    """Tests that set_strategy_stats works as intended."""
    self.strategies.use_corpus_subset = True
    self.strategies.corpus_subset_size = 75
    # Implicitly calls set_strategy_stats
    actual_stats = self._set_stats()
    self.assertEqual(
        actual_stats['strategy_' + strategy.CORPUS_SUBSET_STRATEGY.name], 75)

    # Test that stats for generator strategies are correct.
    self.strategies.generator_strategy = engine_common.Generator.RADAMSA
    actual_stats = self._set_stats()
    self.assertEqual(
        actual_stats['strategy_' +
                     strategy.CORPUS_MUTATION_RADAMSA_STRATEGY.name], 1)

    self.strategies.generator_strategy = engine_common.Generator.ML_RNN
    actual_stats = self._set_stats()
    self.assertEqual(
        actual_stats['strategy_' +
                     strategy.CORPUS_MUTATION_ML_RNN_STRATEGY.name], 1)

  def test_set_output_stats_bad_instrumentation(self):
    """Tests that set_output_stats sets bad_instrumentation properly."""
    stdout_path = os.path.join(self.data_dir, 'bad_instrumentation.txt')
    with open(stdout_path) as file_handle:
      afl_fuzz_output = file_handle.read()
    self.stats_getter.set_output_stats(afl_fuzz_output)
    self.assertEqual(self.stats_getter.stats['bad_instrumentation'], 1)

  def test_set_output_stats_corpus_crash(self):
    """Tests that set_output_stats sets corpus_crash_count properly."""
    stdout_path = os.path.join(self.data_dir, 'corpus_crash.txt')
    with open(stdout_path) as file_handle:
      afl_fuzz_output = file_handle.read()
    self.stats_getter.set_output_stats(afl_fuzz_output)
    self.assertEqual(self.stats_getter.stats['corpus_crash_count'], 1)

  def test_set_output_stats_startup_crash(self):
    """Tests that set_output_stats sets startup_crash_count properly."""
    stdout_path = os.path.join(self.data_dir, 'startup_crash.txt')
    with open(stdout_path) as file_handle:
      afl_fuzz_output = file_handle.read()
    self.stats_getter.set_output_stats(afl_fuzz_output)
    self.assertEqual(self.stats_getter.stats['startup_crash_count'], 1)

  def test_set_output_stats_startup_crash2(self):
    """Tests that set_output_stats sets startup_crash_count properly."""
    stdout_path = os.path.join(self.data_dir, 'startup_crash2.txt')
    with open(stdout_path) as file_handle:
      afl_fuzz_output = file_handle.read()
    self.stats_getter.set_output_stats(afl_fuzz_output)
    self.assertEqual(self.stats_getter.stats['startup_crash_count'], 1)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Profiling functions."""
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment


def start_if_needed(service):
  """Start Google Cloud Profiler if |USE_PYTHON_PROFILER| environment variable
  is set."""
  if not environment.get_value('USE_PYTHON_PROFILER'):
    return True

  project_id = utils.get_application_id()
  service_with_platform = '{service}_{platform}'.format(
      service=service, platform=environment.platform().lower())

  try:
    # Import the package here since it is only needed when profiler is enabled.
    # Also, this is supported on Linux only.
    import googlecloudprofiler
    googlecloudprofiler.start(
        project_id=project_id, service=service_with_platform)
  except Exception:
    logs.log_error(
        'Failed to start the profiler for service %s.' % service_with_platform)
    return False

  return True



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""run_server.py run the Clusterfuzz server locally."""
import os
import shutil
import threading
import time
import urllib.request

from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.tests.test_libs import test_utils
from local.butler import appengine
from local.butler import common
from local.butler import constants


def bootstrap_db():
  """Bootstrap the DB."""

  def bootstrap():
    # Wait for the server to run.
    time.sleep(10)
    print('Bootstrapping datastore...')
    common.execute(
        ('python butler.py run setup '
         '--non-dry-run --local --config-dir={config_dir}'
        ).format(config_dir=constants.TEST_CONFIG_DIR),
        exit_on_error=False)

  thread = threading.Thread(target=bootstrap)
  thread.start()


def create_local_bucket(local_gcs_buckets_path, name):
  """Create a local bucket."""
  blobs_bucket = os.path.join(local_gcs_buckets_path, name)
  if not os.path.exists(blobs_bucket):
    os.mkdir(blobs_bucket)


def bootstrap_gcs(storage_path):
  """Bootstrap GCS."""
  local_gcs_buckets_path = os.path.join(storage_path, 'local_gcs')
  if not os.path.exists(local_gcs_buckets_path):
    os.mkdir(local_gcs_buckets_path)

  config = local_config.ProjectConfig()
  test_blobs_bucket = os.environ.get('TEST_BLOBS_BUCKET')
  if test_blobs_bucket:
    create_local_bucket(local_gcs_buckets_path, test_blobs_bucket)
  else:
    create_local_bucket(local_gcs_buckets_path, config.get('blobs.bucket'))

  create_local_bucket(local_gcs_buckets_path, config.get('deployment.bucket'))
  create_local_bucket(local_gcs_buckets_path, config.get('bigquery.bucket'))
  create_local_bucket(local_gcs_buckets_path, config.get('backup.bucket'))
  create_local_bucket(local_gcs_buckets_path, config.get('logs.fuzzer.bucket'))
  create_local_bucket(local_gcs_buckets_path, config.get('env.CORPUS_BUCKET'))
  create_local_bucket(local_gcs_buckets_path,
                      config.get('env.QUARANTINE_BUCKET'))
  create_local_bucket(local_gcs_buckets_path,
                      config.get('env.SHARED_CORPUS_BUCKET'))
  create_local_bucket(local_gcs_buckets_path,
                      config.get('env.FUZZ_LOGS_BUCKET'))
  create_local_bucket(local_gcs_buckets_path,
                      config.get('env.MUTATOR_PLUGINS_BUCKET'))

  # Symlink local GCS bucket path to appengine src dir to bypass sandboxing
  # issues.
  common.symlink(
      src=local_gcs_buckets_path,
      target=os.path.join(appengine.SRC_DIR_PY, 'local_gcs'))


def start_cron_threads():
  """Start threads to trigger essential cron jobs."""

  request_timeout = 10 * 60  # 10 minutes.

  def trigger(interval_seconds, target):
    """Trigger a cron job."""
    while True:
      time.sleep(interval_seconds)

      try:
        url = 'http://{host}/{target}'.format(
            host=constants.CRON_SERVICE_HOST, target=target)
        request = urllib.request.Request(url)
        request.add_header('X-Appengine-Cron', 'true')
        response = urllib.request.urlopen(request, timeout=request_timeout)
        response.read(60)  # wait for request to finish.
      except Exception:
        continue

  crons = (
      (90, 'cleanup'),
      (60, 'triage'),
      (6 * 3600, 'schedule-progression-tasks'),
      (12 * 3600, 'schedule-corpus-pruning'),
  )

  for interval, cron in crons:
    thread = threading.Thread(target=trigger, args=(interval, cron))
    thread.daemon = True
    thread.start()


def execute(args):
  """Run the server."""
  os.environ['LOCAL_DEVELOPMENT'] = 'True'
  common.kill_leftover_emulators()

  if not args.skip_install_deps:
    common.install_dependencies()

  # Do this everytime as a past deployment might have changed these.
  appengine.symlink_dirs()

  # Deploy all yaml files from test project for basic appengine deployment and
  # local testing to work. This needs to be called on every iteration as a past
  # deployment might have overwritten or deleted these config files.
  yaml_paths = local_config.GAEConfig().get_absolute_path('deployment.prod3')
  appengine.copy_yamls_and_preprocess(yaml_paths)

  # Build templates.
  appengine.build_templates()

  # Clean storage directory if needed.
  if args.bootstrap or args.clean:
    if os.path.exists(args.storage_path):
      print('Clearing local datastore by removing %s.' % args.storage_path)
      shutil.rmtree(args.storage_path)
  if not os.path.exists(args.storage_path):
    os.makedirs(args.storage_path)

  # Set up local GCS buckets and symlinks.
  bootstrap_gcs(args.storage_path)

  # Start pubsub emulator.
  pubsub_emulator = test_utils.start_cloud_emulator(
      'pubsub',
      args=['--host-port=' + constants.PUBSUB_EMULATOR_HOST],
      data_dir=args.storage_path)
  test_utils.setup_pubsub(constants.TEST_APP_ID)

  # Start Datastore emulator
  datastore_emulator = test_utils.start_cloud_emulator(
      'datastore',
      args=['--host-port=' + constants.DATASTORE_EMULATOR_HOST],
      data_dir=args.storage_path,
      store_on_disk=True)

  # Start our custom GCS emulator.
  local_gcs = common.execute_async(
      'go run emulators/gcs.go -storage-path=' + os.path.join(
          os.path.abspath(args.storage_path), 'local_gcs'),
      cwd='local')

  if args.bootstrap:
    bootstrap_db()

  start_cron_threads()

  os.environ['APPLICATION_ID'] = constants.TEST_APP_ID
  os.environ['LOCAL_DEVELOPMENT'] = 'True'
  os.environ['LOCAL_GCS_BUCKETS_PATH'] = 'local_gcs'
  os.environ['LOCAL_GCS_SERVER_HOST'] = constants.LOCAL_GCS_SERVER_HOST
  os.environ['DATASTORE_EMULATOR_HOST'] = constants.DATASTORE_EMULATOR_HOST
  os.environ['PUBSUB_EMULATOR_HOST'] = constants.PUBSUB_EMULATOR_HOST
  os.environ['GAE_ENV'] = 'dev'
  try:
    cron_server = common.execute_async(
        'gunicorn -b :{port} main:app'.format(port=constants.CRON_SERVICE_PORT),
        cwd=os.path.join('src', 'appengine'))

    common.execute(
        'gunicorn -b :{port} main:app'.format(
            port=constants.DEV_APPSERVER_PORT),
        cwd=os.path.join('src', 'appengine'))
  except KeyboardInterrupt:
    print('Server has been stopped. Exit.')
    cron_server.terminate()
    datastore_emulator.cleanup()
    pubsub_emulator.cleanup()
    local_gcs.terminate()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handlers for the bot list."""

import datetime

from flask import request

from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from handlers import base_handler
from libs import filters
from libs import handler
from libs import helpers
from libs.query import datastore_query

PAGE_SIZE = 10
MORE_LIMIT = 50 - PAGE_SIZE  # exactly 5 pages

FILTERS = [
    filters.Keyword([], 'keywords', 'q'),
]


def _get_alive_cutoff():
  """Get the time before which we consider bots to be dead."""
  seconds_to_wait_for_dead_bot = (
      tasks.TASK_LEASE_SECONDS + tasks.TASK_COMPLETION_BUFFER +
      data_types.HEARTBEAT_WAIT_INTERVAL)
  alive_cutoff = utils.utcnow() - datetime.timedelta(
      seconds=seconds_to_wait_for_dead_bot)
  return alive_cutoff


def _convert_heartbeats_to_dicts(heartbeats):
  """Format heartbeats for template."""
  alive_cutoff = _get_alive_cutoff()
  result = []
  for heartbeat in heartbeats:
    result.append({
        'bot_name':
            heartbeat.bot_name,
        'source_version':
            heartbeat.source_version,
        'task_payload':
            heartbeat.task_payload,
        'platform_id':
            heartbeat.platform_id,
        'task_end_time':
            utils.utc_datetime_to_timestamp(heartbeat.task_end_time)
            if heartbeat.task_end_time else '',
        'last_beat_time':
            utils.utc_datetime_to_timestamp(heartbeat.last_beat_time)
            if heartbeat.last_beat_time else '',
        'alive':
            'alive' if heartbeat.last_beat_time > alive_cutoff else 'dead'
    })

  return result


def get_results():
  """Get results for the bots page."""
  # Return bots sorted alphabetically by bot_name
  query = datastore_query.Query(data_types.Heartbeat)
  query.order('bot_name', is_desc=False)
  params = dict(request.iterparams())
  filters.add(query, params, FILTERS)

  page = helpers.cast(request.get('page', 1), int, "'page' is not an int.")
  items, total_pages, total_items, has_more = query.fetch_page(
      page=page, page_size=PAGE_SIZE, projection=None, more_limit=MORE_LIMIT)
  items = _convert_heartbeats_to_dicts(items)
  helpers.log('Bots', helpers.VIEW_OPERATION)

  result = {
      'hasMore': has_more,
      'items': items,
      'page': page,
      'pageSize': PAGE_SIZE,
      'totalItems': total_items,
      'totalPages': total_pages,
  }
  return result, params


class Handler(base_handler.Handler):
  """Handler that gets the bot list."""

  @handler.get(handler.HTML)
  @handler.check_admin_access_if_oss_fuzz
  @handler.check_user_access(need_privileged_access=False)
  def get(self):
    """Render the bot list HTML."""
    result, params = get_results()
    return self.render('bots.html', {
        'result': result,
        'params': params,
    })


class JsonHandler(base_handler.Handler):
  """Handler that gets the bots when user clicks on next page."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.check_admin_access_if_oss_fuzz
  @handler.check_user_access(need_privileged_access=False)
  def post(self):
    """Get and render the bots in JSON."""
    result, _ = get_results()
    return self.render_json(result)


class DeadBotsHandler(base_handler.Handler):
  """Output dead bots as json."""

  @handler.get(handler.JSON)
  def get(self):
    """Render dead bots as json (used by automated scripts)."""

    # This a publicly exposed chromium-specific page.
    if utils.is_chromium():
      heartbeats = ndb_utils.get_all_from_model(data_types.Heartbeat)
    else:
      raise helpers.EarlyExitException('Dead bots list unavailable.', 400)

    result = {}
    alive_cutoff = _get_alive_cutoff()
    for heartbeat in heartbeats:
      if heartbeat.last_beat_time <= alive_cutoff:
        result[heartbeat.bot_name] = 'dead'

    return self.render_json(result)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for recurring_tasks."""

import unittest

import flask
import mock
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import recurring_tasks


@test_utils.with_cloud_emulators('datastore')
class OpenReproducibleTestcaseTasksSchedulerTest(unittest.TestCase):
  """Tests OpenReproducibleTestcaseTasksScheduler."""

  def setUp(self):
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/schedule-open-reproducible-testcase-tasks',
        view_func=recurring_tasks.OpenReproducibleTestcaseTasksScheduler.
        as_view('/schedule-open-reproducible-testcase-tasks'))
    self.app = webtest.TestApp(flaskapp)

    self.testcase_0 = data_types.Testcase(
        open=True,
        one_time_crasher_flag=False,
        status='Processed',
        job_type='job',
        queue='jobs-linux')
    self.testcase_0.put()

    self.testcase_1 = data_types.Testcase(
        open=False,
        one_time_crasher_flag=False,
        status='Processed',
        job_type='job',
        queue='jobs-linux')
    self.testcase_1.put()

    self.testcase_2 = data_types.Testcase(
        open=True,
        one_time_crasher_flag=True,
        status='Processed',
        job_type='job',
        queue='jobs-linux')
    self.testcase_2.put()

    self.testcase_3 = data_types.Testcase(
        open=True,
        one_time_crasher_flag=False,
        status='NA',
        job_type='job',
        queue='jobs-linux')
    self.testcase_3.put()

    self.testcase_4 = data_types.Testcase(
        open=True,
        one_time_crasher_flag=False,
        status='Processed',
        job_type='job_windows',
        queue='jobs-windows')
    self.testcase_4.put()

    data_types.Job(name='job', environment_string='', platform='LINUX').put()
    data_types.Job(
        name='job_windows', environment_string='', platform='WINDOWS').put()

    helpers.patch(self, [
        'handlers.base_handler.Handler.is_cron',
    ])

  def test_execute(self):
    """Tests that we don't directly use this scheduler."""
    with self.assertRaises(webtest.AppError):
      self.app.get('/schedule-open-reproducible-testcase-tasks')


class ProgressionTasksSchedulerTest(OpenReproducibleTestcaseTasksSchedulerTest):
  """Tests ProgressionTasksScheduler."""

  def setUp(self):
    super(ProgressionTasksSchedulerTest, self).setUp()
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/schedule-progression-tasks',
        view_func=recurring_tasks.ProgressionTasksScheduler.as_view(
            '/schedule-progression-tasks'))
    self.app = webtest.TestApp(flaskapp)

    helpers.patch(self, [
        'clusterfuzz._internal.base.tasks.add_task',
    ])

  def test_execute(self):
    """Tests scheduling of progression tasks."""
    self.app.get('/schedule-progression-tasks')
    self.mock.add_task.assert_has_calls([
        mock.call('progression', 1, 'job', queue='jobs-linux'),
        mock.call('progression', 5, 'job_windows', queue='jobs-windows')
    ])


class ImpactTasksSchedulerTest(OpenReproducibleTestcaseTasksSchedulerTest):
  """Tests ProgressionTasksScheduler."""

  def setUp(self):
    super(ImpactTasksSchedulerTest, self).setUp()
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/schedule-impact-tasks',
        view_func=recurring_tasks.ImpactTasksScheduler.as_view(
            '/schedule-impact-tasks'))
    self.app = webtest.TestApp(flaskapp)
    helpers.patch(self, [
        'clusterfuzz._internal.base.tasks.add_task',
    ])

  def test_execute(self):
    """Tests scheduling of progression tasks."""
    self.app.get('/schedule-impact-tasks')
    self.mock.add_task.assert_has_calls([
        mock.call('impact', 1, 'job', queue='jobs-linux'),
        mock.call('impact', 5, 'job_windows', queue='jobs-windows'),
    ])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the generic blackbox fuzzer engine implementation."""

import os
import unittest
from unittest import mock

from clusterfuzz._internal.bot.fuzzers.blackbox import engine
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers

TEST_PATH = os.path.abspath(os.path.dirname(__file__))
DATA_DIR = os.path.join(TEST_PATH, 'test_data')
FUZZ_OUTPUT_DIR = os.path.join(DATA_DIR, 'sample_fuzz_output')


class BlackboxEngineTest(unittest.TestCase):
  """Tests for BlackboxEngine."""

  def setUp(self):
    test_helpers.patch_environ(self)
    environment.set_bot_environment()

    os.environ['APP_ARGS'] = '-a -b'
    os.environ['APP_DIR'] = '/build'
    os.environ['APP_PATH'] = '/build/test_binary'
    os.environ['APP_NAME'] = 'test_binary'

    os.environ['FUZZERS_DIR'] = '/fuzzers'

    test_helpers.patch(self, [
        'os.chmod',
        'clusterfuzz._internal.system.new_process.ProcessRunner.run_and_wait',
    ])

  def test_prepare(self):
    blackbox_engine = engine.BlackboxEngine()
    result = blackbox_engine.prepare('/input/corpus', 'unused', '/build')
    self.assertEqual(result.corpus_dir, '/input/corpus')
    self.assertEqual(result.arguments, [])
    self.assertEqual(result.strategies, {})

  def test_fuzz(self):
    blackbox_engine = engine.BlackboxEngine()
    options = blackbox_engine.prepare('/input/corpus', 'unused', '/build')
    result = blackbox_engine.fuzz('/fuzzers/my_fuzzer', options,
                                  FUZZ_OUTPUT_DIR, 10)
    self.assertEqual(len(result.crashes), 1)
    self.assertEqual(result.crashes[0].input_path,
                     os.path.join(FUZZ_OUTPUT_DIR, 'fuzz-real-crash'))
    self.mock.run_and_wait.assert_called_once_with(
        mock.ANY,
        additional_args=[
            '--app_path=/build/test_binary', '--app_args=-a -b',
            '--input_dir=/input/corpus'
        ],
        timeout=10)

  def test_reproduce(self):
    blackbox_engine = engine.BlackboxEngine()
    args = ['fuzzer_executable']
    blackbox_engine.reproduce('/fuzzers/my_fuzzer', '/testcase', args, 10)
    self.mock.run_and_wait.assert_called_once_with(
        mock.ANY,
        additional_args=[
            '--app_path=/build/test_binary', '--app_args=-a -b',
            '--testcase_path=/testcase'
        ],
        timeout=10)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manage automatic weight adjustments."""

import collections
import datetime

from google.cloud import ndb
import six

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.metrics import fuzzer_stats
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from handlers import base_handler
from libs import handler

QuerySpecification = collections.namedtuple(
    'QuerySpecification', ['query_format', 'formatter', 'reason'])

SpecificationMatch = collections.namedtuple('SpecificationMatch',
                                            ['new_weight', 'reason'])

DEFAULT_MULTIPLIER = 30.0  # Used for blackbox and jobs that are not yet run.
DEFAULT_SANITIZER_WEIGHT = 0.1
DEFAULT_ENGINE_WEIGHT = 1.0
TARGET_COUNT_WEIGHT_CAP = 100.0

SANITIZER_BASE_WEIGHT = 0.1

# TODO(ochang): architecture weights.
SANITIZER_WEIGHTS = {
    'ASAN': 5 * SANITIZER_BASE_WEIGHT,
    'CFI': 1 * SANITIZER_BASE_WEIGHT,
    'MSAN': 2 * SANITIZER_BASE_WEIGHT,
    'TSAN': 1 * SANITIZER_BASE_WEIGHT,
    'UBSAN': 1 * SANITIZER_BASE_WEIGHT,
}

ENGINE_WEIGHTS = {
    'libFuzzer': 1.0,
    'afl': 1.0,
    'honggfuzz': 0.2,
}


# Formatters for query specifications.
def _past_day_formatter(query_format, dataset):
  """Simple formatter to get stats for the past day."""
  end_time = utils.utcnow().date()
  start_time = end_time - datetime.timedelta(days=1)
  return query_format.format(
      dataset=dataset, start_time=start_time, end_time=end_time)


def _new_fuzzer_formatter(query_format, dataset):
  """Prepare a query to check for new fuzzers from the past week."""
  now = utils.utcnow().date()
  cutoff_time = now - datetime.timedelta(days=7)
  return query_format.format(dataset=dataset, cutoff_time=cutoff_time)


def _coverage_formatter(query_format, dataset):
  """Prepare a query to check for changes in coverage week over week."""
  end_date = utils.utcnow().date() - datetime.timedelta(days=1)
  middle_date = end_date - datetime.timedelta(days=7)
  start_date = end_date - datetime.timedelta(days=14)
  return query_format.format(
      dataset=dataset,
      start_date=start_date,
      middle_date=middle_date,
      end_date=end_date)


# Most of our queries should simply average a field name to get a ratio showing
# how often some behavior occurs.
GENERIC_QUERY_FORMAT = """
SELECT
  fuzzer,
  job,
  1.0 - (1.0 - {min_weight}) * AVG({field_name}) AS new_weight
FROM
  {{dataset}}.TestcaseRun
WHERE
  _PARTITIONTIME BETWEEN TIMESTAMP('{{start_time}}')
  AND TIMESTAMP('{{end_time}}')
GROUP BY
  fuzzer,
  job
"""

# Heavily reduce the weight for fuzzers which frequently crash on startup. This
# is indicitave of a very serious problem that makes it highly unlikely that
# we'll find anything during fuzzing.
STARTUP_CRASH_SPECIFICATION = QuerySpecification(
    query_format=GENERIC_QUERY_FORMAT.format(
        field_name='startup_crash_count', min_weight=0.10),
    formatter=_past_day_formatter,
    reason='frequent startup crashes')

# Reduce weight somewhat for fuzzers with many slow units. If a particular unit
# runs for so long that we detect it as a slow unit, it usually means that the
# fuzzer is not making good use of its cycles while running or needs a fix.
SLOW_UNIT_SPECIFICATION = QuerySpecification(
    query_format=GENERIC_QUERY_FORMAT.format(
        field_name='slow_unit_count', min_weight=0.25),
    formatter=_past_day_formatter,
    reason='frequent slow units')

# This should end up being very similar to the slow unit specification, and is
# included for the same reason.
TIMEOUT_SPECIFICATION = QuerySpecification(
    query_format=GENERIC_QUERY_FORMAT.format(
        field_name='timeout_count', min_weight=0.25),
    formatter=_past_day_formatter,
    reason='frequent timeouts')

# Fuzzers with extremely frequent OOMs may contain leaks or other issues that
# signal that they need some improvement. Run with a slightly reduced weight
# until the issues are fixed.
OOM_SPECIFICATION = QuerySpecification(
    query_format=GENERIC_QUERY_FORMAT.format(
        field_name='oom_count', min_weight=0.25),
    formatter=_past_day_formatter,
    reason='frequent OOMs')

# Fuzzers which are crashing frequently may not be making full use of their
# allotted time for fuzzing, and may end up being more effective once the known
# issues are fixed. This rule is more lenient than some of the others as even
# healthy fuzzers are expected to have some crashes.
CRASH_SPECIFICATION = QuerySpecification(
    query_format=GENERIC_QUERY_FORMAT.format(
        field_name='crash_count', min_weight=0.50),
    formatter=_past_day_formatter,
    reason='frequent crashes')

# New fuzzers/jobs should run much more frequently than others. In this case, we
# test the fraction of days for which we have no stats for this fuzzer/job pair
# and increase if it's nonzero.
NEW_FUZZER_FORMAT = """
SELECT
  fuzzer,
  job,
  5.0 as new_weight,
  MIN(_PARTITIONTIME) as first_time
FROM
  {dataset}.TestcaseRun
GROUP BY
  fuzzer,
  job
HAVING
  first_time >= TIMESTAMP('{cutoff_time}')
"""

NEW_FUZZER_SPECIFICATION = QuerySpecification(
    query_format=NEW_FUZZER_FORMAT,
    formatter=_new_fuzzer_formatter,
    reason='new fuzzer')

# Format to query for fuzzers with minimal change in week to week coverage.
COVERAGE_UNCHANGED_FORMAT = """
SELECT
  recent.fuzzer AS fuzzer,
  recent.job AS job,
  0.75 as new_weight
FROM (
  SELECT
    fuzzer,
    job,
    MAX(edge_coverage / edges_total) AS coverage
  FROM
    {dataset}.TestcaseRun
  WHERE
    _PARTITIONTIME BETWEEN TIMESTAMP('{middle_date}')
    AND TIMESTAMP('{end_date}')
    AND edges_total > 0
    AND edge_coverage > 0
  GROUP BY
    fuzzer,
    job
  HAVING
    coverage <= 1.0) AS recent
JOIN (
  SELECT
    fuzzer,
    job,
    MAX(edge_coverage / edges_total) AS coverage
  FROM
    {dataset}.TestcaseRun
  WHERE
    _PARTITIONTIME BETWEEN TIMESTAMP('{start_date}')
    AND TIMESTAMP('{middle_date}')
    AND edges_total > 0
    AND edge_coverage > 0
  GROUP BY
    fuzzer,
    job
  HAVING
    coverage <= 1.0) AS older
ON
  recent.fuzzer = older.fuzzer
  AND recent.job = older.job
WHERE
  ABS((recent.coverage - older.coverage) / recent.coverage) < 0.01
"""

COVERAGE_UNCHANGED_SPECIFICATION = QuerySpecification(
    query_format=COVERAGE_UNCHANGED_FORMAT,
    formatter=_coverage_formatter,
    reason='coverage flat over past 2 weeks')

# Mappings for which specifications to use for which
LIBFUZZER_SPECIFICATIONS = [
    COVERAGE_UNCHANGED_SPECIFICATION,
    CRASH_SPECIFICATION,
    NEW_FUZZER_SPECIFICATION,
    OOM_SPECIFICATION,
    SLOW_UNIT_SPECIFICATION,
    STARTUP_CRASH_SPECIFICATION,
    TIMEOUT_SPECIFICATION,
]

AFL_SPECIFICATIONS = [
    CRASH_SPECIFICATION,
    NEW_FUZZER_SPECIFICATION,
    STARTUP_CRASH_SPECIFICATION,
]

RESTORE_DEFAULT_MATCH = SpecificationMatch(
    new_weight=1.0, reason='no longer matches any weight adjustment rules')


def _query_helper(client, query):
  """Helper function to get fuzzer stats."""
  return client.query(query=query).rows


def _update_match(matches, fuzzer, job, match):
  """Update the weight for a fuzzer/job."""
  key = (fuzzer, job)
  old_match = matches.get(key, RESTORE_DEFAULT_MATCH)

  new_weight = match.new_weight
  old_weight = old_match.new_weight

  # Rules that increase weights are expected to take precedence over any that
  # lower the weight. Issues with new fuzzers may be fixed intraday and other
  # issues like crashes shouldn't be penalized for them.
  if old_weight > 1.0:
    return

  # Always update the weight if the previous value is the default. This is
  # required to deal with specifications that are meant to set the weight above
  # 1.0. Otherwise, prioritize only the most penalizing match for this pairing.
  if old_match == RESTORE_DEFAULT_MATCH or new_weight < old_weight:
    matches[key] = match


def update_weight_for_target(fuzz_target_name, job, match):
  """Set the weight for a particular target."""
  target_job = data_handler.get_fuzz_target_job(fuzz_target_name, job)

  if not target_job:
    # Bail out. This is expected if any fuzzer/job combinations become outdated.
    return

  weight = match.new_weight
  logs.log('Adjusted weight to %f for target %s and job %s (%s).' %
           (weight, fuzz_target_name, job, match.reason))

  target_job.weight = weight
  target_job.put()


def update_matches_for_specification(specification, client, engine, matches,
                                     run_set):
  """Run a query and adjust weights based on a given query specification."""
  query = specification.formatter(specification.query_format,
                                  fuzzer_stats.dataset_name(engine))
  results = _query_helper(client, query)
  for result in results:
    fuzzer = result['fuzzer']
    job = result['job']
    new_weight = result['new_weight']

    if new_weight is None:
      continue

    run_set.add((fuzzer, job))
    if new_weight != 1.0:
      match = SpecificationMatch(
          new_weight=new_weight, reason=specification.reason)
      _update_match(matches, fuzzer, job, match)


def update_target_weights_for_engine(client, engine, specifications):
  """Update all fuzz target weights for the specified engine."""
  matches = {}
  run_set = set()

  # All fuzzers with non-default weights must be tracked with a special
  # specification. This ensures that they will be restored to normal weight
  # once conditions causing adjustments are no longer met.
  target_jobs = data_types.FuzzTargetJob.query(
      data_types.FuzzTarget.engine == engine).filter(
          data_types.FuzzTargetJob.weight != 1.0)

  for target_job in target_jobs:
    matches[(target_job.fuzz_target_name,
             target_job.job)] = RESTORE_DEFAULT_MATCH

  for match in specifications:
    update_matches_for_specification(match, client, engine, matches, run_set)

  for (fuzzer, job), match in six.iteritems(matches):
    if (fuzzer, job) not in run_set:
      # This ensures that we don't reset weights for fuzzers with problems if
      # they didn't run in the time covered by our queries.
      continue

    update_weight_for_target(fuzzer, job, match)

  logs.log('Weight adjustments complete for engine %s.' % engine)


def store_current_weights_in_bigquery():
  """Update a bigquery table containing the daily stats."""
  rows = []
  target_jobs = ndb_utils.get_all_from_model(data_types.FuzzTargetJob)
  for target_job in target_jobs:
    row = {
        'fuzzer': target_job.fuzz_target_name,
        'job': target_job.job,
        'weight': target_job.weight
    }
    rows.append(big_query.Insert(row=row, insert_id=None))

  client = big_query.Client(dataset_id='main', table_id='fuzzer_weights')
  client.insert(rows)


def update_job_weight(job_name, multiplier):
  """Update a job weight."""
  tool_name = environment.get_memory_tool_name(job_name)
  multiplier *= SANITIZER_WEIGHTS.get(tool_name, DEFAULT_SANITIZER_WEIGHT)

  engine = environment.get_engine_for_job(job_name)
  multiplier *= ENGINE_WEIGHTS.get(engine, DEFAULT_ENGINE_WEIGHT)

  query = data_types.FuzzerJob.query(data_types.FuzzerJob.job == job_name)
  changed_weights = []
  for fuzzer_job in query:
    if fuzzer_job.multiplier != multiplier:
      fuzzer_job.multiplier = multiplier
      changed_weights.append(fuzzer_job)

  if changed_weights:
    ndb_utils.put_multi(changed_weights)


def update_job_weights():
  """Update job weights."""
  for job in data_types.Job.query():
    multiplier = DEFAULT_MULTIPLIER
    if environment.is_engine_fuzzer_job(job.name):
      targets_count = ndb.Key(data_types.FuzzTargetsCount, job.name).get()
      # If the count is 0, it may be due to a bad build or some other issue. Use
      # the default weight in that case to allow for recovery.
      if targets_count and targets_count.count:
        multiplier = targets_count.count
        multiplier = min(multiplier, TARGET_COUNT_WEIGHT_CAP)

    update_job_weight(job.name, multiplier)


class Handler(base_handler.Handler):
  """Handler to periodically update fuzz target weights based on performance."""

  @handler.cron()
  def get(self):
    """Process all fuzz targets and update FuzzTargetJob weights."""
    client = big_query.Client()
    update_target_weights_for_engine(client, 'libFuzzer',
                                     LIBFUZZER_SPECIFICATIONS)
    update_target_weights_for_engine(client, 'afl', AFL_SPECIFICATIONS)
    update_job_weights()

    store_current_weights_in_bigquery()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""App Engine GCS helpers."""

import base64
import collections
import datetime
import json
import time
import urllib.parse

import googleapiclient

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.system import environment

STORAGE_URL = 'https://storage.googleapis.com/%s'
DEFAULT_URL_VALID_SECONDS = 30 * 60  # 30 minutes.
MAX_UPLOAD_SIZE = 15 * 1024 * 1024 * 1024  # 15 GB.

GcsUpload = collections.namedtuple(
    'GcsUpload',
    ['url', 'bucket', 'key', 'google_access_id', 'policy', 'signature'])


class GcsError(Exception):
  """Base class for exceptions in this module."""


def sign_data(data):
  """Sign data with the default App Engine service account."""
  iam = googleapiclient.discovery.build('iamcredentials', 'v1')
  service_account = 'projects/-/serviceAccounts/' + utils.service_account_email(
  )

  response = iam.projects().serviceAccounts().signBlob(
      name=service_account,
      body={
          'delegates': [],
          'payload': base64.b64encode(data).decode('utf-8'),
      }).execute()

  try:
    return base64.b64decode(response['signedBlob'])
  except Exception as e:
    raise GcsError('Invalid response: ' + str(e))


class SignedGcsHandler(object):
  """Handler for signing and redirecting to a GCS object."""

  def serve_gcs_object(self, bucket, object_path, content_disposition=None):
    """Serve a GCS object."""
    url = get_signed_url(bucket, object_path)

    if content_disposition:
      content_disposition_params = {
          'response-content-disposition': content_disposition,
      }

      url += '&' + urllib.parse.urlencode(content_disposition_params)

    return self.redirect(url)


def _get_expiration_time(expiry_seconds):
  """Return a timestamp |expiry_seconds| from now."""
  return int(time.time() + expiry_seconds)


def get_signed_url(bucket_name,
                   path,
                   method='GET',
                   expiry=DEFAULT_URL_VALID_SECONDS):
  """Return a signed url."""
  timestamp = _get_expiration_time(expiry)
  blob = '%s\n\n\n%d\n/%s/%s' % (method, timestamp, bucket_name, path)

  local_server = environment.get_value('LOCAL_GCS_SERVER_HOST')
  if local_server:
    url = local_server + '/' + bucket_name
    signed_blob = b'SIGNATURE'
    service_account_name = 'service_account'
  else:
    url = STORAGE_URL % bucket_name
    signed_blob = sign_data(blob.encode('utf-8'))
    service_account_name = utils.service_account_email()

  params = {
      'GoogleAccessId': service_account_name,
      'Expires': timestamp,
      'Signature': base64.b64encode(signed_blob).decode('utf-8'),
  }

  return str(url + '/' + path + '?' + urllib.parse.urlencode(params))


def prepare_upload(bucket_name, path, expiry=DEFAULT_URL_VALID_SECONDS):
  """Prepare a signed GCS upload."""
  expiration_time = (
      datetime.datetime.utcnow() + datetime.timedelta(seconds=expiry))

  conditions = [
      {
          'key': path
      },
      {
          'bucket': bucket_name
      },
      ['content-length-range', 0, MAX_UPLOAD_SIZE],
      ['starts-with', '$x-goog-meta-filename', ''],
  ]

  policy = base64.b64encode(
      json.dumps({
          'expiration': expiration_time.isoformat() + 'Z',
          'conditions': conditions,
      }).encode('utf-8'))

  local_server = environment.get_value('LOCAL_GCS_SERVER_HOST')
  if local_server:
    url = local_server
    signature = b'SIGNATURE'
    service_account_name = 'service_account'
  else:
    url = STORAGE_URL % bucket_name
    signature = base64.b64encode(sign_data(policy))
    service_account_name = utils.service_account_email()

  return GcsUpload(url, bucket_name, path, service_account_name, policy,
                   signature)


def prepare_blob_upload():
  """Prepare a signed GCS blob upload."""
  return prepare_upload(storage.blobs_bucket(), blobs.generate_new_blob_name())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Windows initialization script."""

import unittest

import mock

from clusterfuzz._internal.bot.init_scripts import windows
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers


class CleanTempDirectoriesTest(unittest.TestCase):
  """Test clean_temp_directories."""

  def setUp(self):
    helpers.patch(self, [
        'os.path.abspath',
        'os.path.expandvars',
        'os.path.join',
        'clusterfuzz._internal.system.shell.remove_directory',
    ])

    def abspath(path):
      return path

    def expandvars(path):
      path = path.replace('%TEMP%', r'C:\Users\clusterfuzz\AppData\Local\Temp')
      path = path.replace('%USERPROFILE%', r'C:\Users\clusterfuzz')
      path = path.replace('%WINDIR%', r'C:\WINDOWS')
      return path

    def join(path1, path2):
      """Windows specific os.path.join"""
      return r'%s\%s' % (path1.rstrip('\\'), path2)

    self.mock.abspath.side_effect = abspath
    self.mock.expandvars.side_effect = expandvars
    self.mock.join.side_effect = join

  def test(self):
    windows.clean_temp_directories()

    self.mock.remove_directory.assert_has_calls([
        mock.call(
            r'C:\Users\clusterfuzz\AppData\Local\Temp',
            recreate=True,
            ignore_errors=True),
        mock.call(
            r'C:\Users\clusterfuzz\AppVerifierLogs',
            recreate=True,
            ignore_errors=True),
        mock.call(
            r'C:\Users\clusterfuzz\Downloads',
            recreate=True,
            ignore_errors=True),
        mock.call(r'C:\WINDOWS\Temp', recreate=True, ignore_errors=True),
        mock.call(
            r'C:\Program Files (x86)\Windows Kits\10\Debuggers\x86\sym',
            recreate=True,
            ignore_errors=True),
        mock.call(
            r'C:\Program Files (x86)\Windows Kits\10\Debuggers\x64\sym',
            recreate=True,
            ignore_errors=True)
    ])


class RemountIfNeededTest(unittest.TestCase):
  """Test remount_if_needed."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.metrics.logs.log_error',
        'clusterfuzz._internal.base.retry.sleep',
        'clusterfuzz._internal.base.utils.write_data_to_file',
        'os.path.exists',
        'os.path.join',
        'subprocess.call',
        'subprocess.check_call',
    ])

    def join(path1, path2):
      """Windows specific os.path.join"""
      return r'%s\%s' % (path1.rstrip('\\'), path2)

    self.mock.join.side_effect = join

    environment.set_value('NFS_HOST', 'clusterfuzz-windows-0001')
    environment.set_value('NFS_VOLUME', 'cfvolume')
    environment.set_value('NFS_ROOT', 'X:\\')

  def test_with_mount_and_with_check_file(self):
    """Test remount_if_needed when mount works and check file already exists."""
    self.mock.exists.return_value = True
    windows.remount_if_needed()

    self.assertEqual(0, self.mock.call.call_count)
    self.assertEqual(0, self.mock.check_call.call_count)
    self.assertEqual(0, self.mock.write_data_to_file.call_count)

  def test_without_mount_and_without_check_file_no_retry(self):
    """Test remount_if_needed when mount and check file do not exist and gets
    created later on successful remount."""
    self.mock.exists.side_effect = [False, False, True]
    windows.remount_if_needed()

    self.mock.call.assert_called_once_with(['umount', '-f', 'X:\\'])
    self.mock.check_call.assert_called_once_with([
        'mount', '-o', 'anon', '-o', 'nolock', '-o', 'retry=10',
        'clusterfuzz-windows-0001:/cfvolume', 'X:\\'
    ])
    self.mock.write_data_to_file.assert_called_once_with('ok', r'X:\check')

  def test_without_mount_and_with_check_file_no_retry(self):
    """Test remount_if_needed when mount does not exist, but check file does and
    check file does not get recreated later on successful remount."""
    self.mock.exists.side_effect = [False, True]
    windows.remount_if_needed()

    self.mock.call.assert_called_once_with(['umount', '-f', 'X:\\'])
    self.mock.check_call.assert_called_once_with([
        'mount', '-o', 'anon', '-o', 'nolock', '-o', 'retry=10',
        'clusterfuzz-windows-0001:/cfvolume', 'X:\\'
    ])
    self.assertEqual(0, self.mock.write_data_to_file.call_count)

  def test_without_mount_and_without_check_file_retry(self):
    """Test remount_if_needed when check file does not exist and gets created
    later on second remount try."""
    self.mock.exists.side_effect = [False, False, False, False, True]
    windows.remount_if_needed()

    self.mock.call.assert_has_calls([mock.call(['umount', '-f', 'X:\\'])] * 2)
    self.mock.check_call.assert_has_calls([
        mock.call([
            'mount', '-o', 'anon', '-o', 'nolock', '-o', 'retry=10',
            'clusterfuzz-windows-0001:/cfvolume', 'X:\\'
        ])
    ] * 2)
    self.mock.write_data_to_file.assert_called_once_with('ok', r'X:\check')

  def test_without_check_file_fail(self):
    """Test remount_if_needed when check file does not exist and does not get
    recreated due to remount failure."""
    self.mock.exists.side_effect = [False, False, False] * 6

    with self.assertRaises(Exception):
      windows.remount_if_needed()

    self.mock.call.assert_has_calls([mock.call(['umount', '-f', 'X:\\'])] * 6)
    self.mock.check_call.assert_has_calls([
        mock.call([
            'mount', '-o', 'anon', '-o', 'nolock', '-o', 'retry=10',
            'clusterfuzz-windows-0001:/cfvolume', 'X:\\'
        ])
    ] * 6)
    self.mock.write_data_to_file.assert_has_calls(
        [mock.call('ok', r'X:\check')] * 6)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for oss_fuzz_build_status."""

import datetime
import json
import unittest

import flask
import mock
import six
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import oss_fuzz_build_status
from libs.issue_management import monorail
from libs.issue_management.monorail.issue import Issue


class MockResponse(object):
  """Mock url request's response."""

  def __init__(self, text):
    self.text = text

  def raise_for_status(self):
    pass


class IssueTrackerManager(object):
  """Mock issue tracker manager."""

  def __init__(self, project_name):
    self.project_name = project_name
    self.issues = {}
    self.next_id = 1

  def get_issue(self, issue_id):
    """Get original issue."""
    issue = self.issues[issue_id]
    issue.itm = self
    return issue

  def save(self, issue, *args, **kwargs):  # pylint: disable=unused-argument
    """Save an issue."""
    if issue.new:
      issue.id = self.next_id
      self.next_id += 1

    self.issues[issue.id] = issue


@test_utils.with_cloud_emulators('datastore')
class OssFuzzBuildStatusTest(unittest.TestCase):
  """Tests for oss_fuzz_build_status."""

  def setUp(self):
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/build-status',
        view_func=oss_fuzz_build_status.Handler.as_view('/build-status'))
    self.app = webtest.TestApp(flaskapp)

    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
        'handlers.base_handler.Handler.is_cron',
        'libs.issue_management.issue_tracker_utils.get_issue_tracker',
        'clusterfuzz._internal.metrics.logs.log_error',
        'requests.get',
    ])

    self.mock.utcnow.return_value = datetime.datetime(2018, 2, 1)
    self.mock.is_cron.return_value = True

    self.itm = IssueTrackerManager('oss-fuzz')
    self.mock.get_issue_tracker.return_value = monorail.IssueTracker(self.itm)

    self.maxDiff = None  # pylint: disable=invalid-name

  def test_no_build_failures(self):
    """Test run with no build failures."""
    # Return the same status for all build types.
    self.mock.get.return_value = MockResponse(
        json.dumps({
            'projects': [
                {
                    'history': [{
                        'finish_time': '2018-02-01T00:00:00.000000Z',
                        'build_id': 'proj0-id',
                        'success': True
                    }],
                    'name':
                        'proj0',
                },
                {
                    'history': [{
                        'finish_time': '2018-02-01T00:00:00.000000Z',
                        'build_id': 'proj1-id',
                        'success': True
                    }],
                    'name':
                        'proj1',
                },
            ]
        }))

    self.app.get('/build-status')
    self.assertEqual(0, data_types.OssFuzzBuildFailure.query().count())
    self.assertEqual(0, len(self.itm.issues))

  def test_build_failures(self):
    """Test run with multiple build failures of different type."""

    def _mock_requests_get(url):
      """Mock requests.get."""
      if url == oss_fuzz_build_status.FUZZING_STATUS_URL:
        return MockResponse(
            json.dumps({
                'projects': [
                    # Both fuzzing and coverage build types are successful.
                    {
                        'history': [{
                            'finish_time': '2018-02-01T00:00:00.000000Z',
                            'build_id': 'proj0-id-f',
                            'success': True
                        }],
                        'name':
                            'proj0',
                    },
                    # Only coverage build type is broken for a while.
                    {
                        'history': [{
                            'finish_time': '2018-02-01T00:00:00.000000Z',
                            'build_id': 'proj5-id-f',
                            'success': True
                        }],
                        'name':
                            'proj5',
                    },
                    # Only coverage build type broken.
                    {
                        'history': [{
                            'finish_time': '2018-02-01T00:00:00.000000Z',
                            'build_id': 'proj6-id-f',
                            'success': True
                        }],
                        'name':
                            'proj6',
                    },

                    # New failure (first 1).
                    {
                        'history': [{
                            'finish_time': '2018-02-01T00:00:00.000000000Z',
                            'build_id': 'proj1-id-f',
                            'success': False
                        }],
                        'name':
                            'proj1',
                    },
                    # Seen failure (second consecutive).
                    {
                        'history': [{
                            'finish_time': '2018-02-01T00:00:00.000000Z',
                            'build_id': 'proj2-id-f',
                            'success': False
                        }],
                        'name':
                            'proj2',
                    },
                    # Seen failure (not updated).
                    {
                        'history': [{
                            'finish_time': '2018-01-31T00:00:00.000000Z',
                            'build_id': 'proj3-id-f',
                            'success': False
                        }],
                        'name':
                            'proj3',
                    },
                    # Seen failure (third consecutive, bug already filed).
                    {
                        'history': [{
                            'finish_time': '2018-02-01T00:00:00.000000Z',
                            'build_id': 'proj4-id-f',
                            'success': False
                        }],
                        'name':
                            'proj4',
                    },
                ]
            }))

      assert url == oss_fuzz_build_status.COVERAGE_STATUS_URL
      return MockResponse(
          json.dumps({
              'projects': [
                  # Both fuzzing and coverage build types are successful.
                  {
                      'history': [{
                          'finish_time': '2018-02-01T00:00:00.000000Z',
                          'build_id': 'proj0-id-c',
                          'success': True
                      }],
                      'name':
                          'proj0',
                  },

                  # New failure (first 1).
                  {
                      'history': [{
                          'finish_time': '2018-02-01T00:00:00.000000000Z',
                          'build_id': 'proj1-id-c',
                          'success': False
                      }],
                      'name':
                          'proj1',
                  },
                  # Seen failure (second consecutive).
                  {
                      'history': [{
                          'name': 'proj2',
                          'finish_time': '2018-02-01T00:00:00.000000Z',
                          'success': False
                      }],
                      'name':
                          'proj2',
                  },
                  # Seen failure (not updated).
                  {
                      'history': [{
                          'finish_time': '2018-01-31T00:00:00.000000Z',
                          'build_id': 'proj3-id-c',
                          'success': False
                      }],
                      'name':
                          'proj3',
                  },
                  # Seen failure (third consecutive, bug already filed).
                  {
                      'history': [{
                          'finish_time': '2018-02-01T00:00:00.000000Z',
                          'build_id': 'proj4-id-c',
                          'success': False
                      }],
                      'name':
                          'proj4',
                  },
                  # Coverage build type is broken for a while.
                  {
                      'history': [{
                          'finish_time': '2018-02-01T00:00:00.000000Z',
                          'build_id': 'proj5-id-c',
                          'success': False
                      }],
                      'name':
                          'proj5',
                  },
                  # Only coverage build type broken (second consecutive).
                  {
                      'history': [{
                          'finish_time': '2018-02-01T00:00:00.000000Z',
                          'build_id': 'proj6-id-c',
                          'success': False
                      }],
                      'name':
                          'proj6',
                  },
              ]
          }))

    self.mock.get.side_effect = _mock_requests_get

    data_types.OssFuzzBuildFailure(
        id='proj2',
        project_name='proj2',
        last_checked_timestamp=datetime.datetime(2018, 1, 31),
        consecutive_failures=1,
        build_type='fuzzing').put()

    data_types.OssFuzzBuildFailure(
        id='proj3',
        project_name='proj3',
        last_checked_timestamp=datetime.datetime(2018, 1, 31),
        consecutive_failures=1,
        build_type='fuzzing').put()

    data_types.OssFuzzBuildFailure(
        id='proj4',
        project_name='proj4',
        last_checked_timestamp=datetime.datetime(2018, 1, 31),
        issue_id='1337',
        consecutive_failures=2,
        build_type='fuzzing').put()

    data_types.OssFuzzBuildFailure(
        id='proj5-coverage',
        project_name='proj5',
        last_checked_timestamp=datetime.datetime(2018, 1, 31),
        issue_id='31337',
        consecutive_failures=5,
        build_type='coverage').put()

    data_types.OssFuzzBuildFailure(
        id='proj6-coverage',
        project_name='proj6',
        last_checked_timestamp=datetime.datetime(2018, 1, 31),
        issue_id=None,
        consecutive_failures=1,
        build_type='coverage').put()

    data_types.OssFuzzProject(
        id='proj2', name='proj2', ccs=['a@user.com']).put()
    data_types.OssFuzzProject(
        id='proj6', name='proj7', ccs=['b@user.com']).put()

    self.app.get('/build-status')
    six.assertCountEqual(self, [
        {
            'build_type': 'fuzzing',
            'consecutive_failures': 1,
            'issue_id': None,
            'last_checked_timestamp': datetime.datetime(2018, 2, 1, 0, 0),
            'project_name': u'proj1'
        },
        {
            'build_type': 'fuzzing',
            'consecutive_failures': 2,
            'issue_id': '1',
            'last_checked_timestamp': datetime.datetime(2018, 2, 1, 0, 0),
            'project_name': u'proj2'
        },
        {
            'build_type': 'fuzzing',
            'consecutive_failures': 1,
            'issue_id': None,
            'last_checked_timestamp': datetime.datetime(2018, 1, 31, 0, 0),
            'project_name': u'proj3'
        },
        {
            'build_type': 'fuzzing',
            'consecutive_failures': 3,
            'issue_id': '1337',
            'last_checked_timestamp': datetime.datetime(2018, 2, 1, 0, 0),
            'project_name': u'proj4'
        },
        {
            'build_type': 'coverage',
            'consecutive_failures': 6,
            'issue_id': '31337',
            'last_checked_timestamp': datetime.datetime(2018, 2, 1, 0, 0),
            'project_name': u'proj5'
        },
        {
            'build_type': 'coverage',
            'consecutive_failures': 2,
            'issue_id': '2',
            'last_checked_timestamp': datetime.datetime(2018, 2, 1, 0, 0),
            'project_name': u'proj6'
        },
    ], [
        failure.to_dict() for failure in data_types.OssFuzzBuildFailure.query()
    ])

    self.assertEqual(2, len(self.itm.issues))
    issue = self.itm.issues[1]
    six.assertCountEqual(self, ['a@user.com'], issue.cc)
    self.assertEqual('New', issue.status)
    self.assertEqual('proj2: Fuzzing build failure', issue.summary)
    self.assertEqual(
        'The last 2 builds for proj2 have been failing.\n'
        '<b>Build log:</b> '
        'https://oss-fuzz-build-logs.storage.googleapis.com/'
        'log-proj2-id-f.txt\n'
        'Build type: fuzzing\n\n'
        'To reproduce locally, please see: '
        'https://google.github.io/oss-fuzz/advanced-topics/reproducing'
        '#reproducing-build-failures\n\n'
        '<b>This bug tracker is not being monitored by OSS-Fuzz team.</b> '
        'If you have any questions, please create an issue at '
        'https://github.com/google/oss-fuzz/issues/new.\n\n'
        '**This bug will be automatically closed within a '
        'day once it is fixed.**', issue.body)

    self.assertTrue(issue.has_label('Proj-proj2'))
    self.assertTrue(issue.has_label('Type-Build-Failure'))

    issue = self.itm.issues[2]
    six.assertCountEqual(self, ['b@user.com'], issue.cc)
    self.assertEqual('New', issue.status)
    self.assertEqual('proj6: Coverage build failure', issue.summary)
    self.assertEqual(
        'The last 2 builds for proj6 have been failing.\n'
        '<b>Build log:</b> '
        'https://oss-fuzz-build-logs.storage.googleapis.com/'
        'log-proj6-id-c.txt\n'
        'Build type: coverage\n\n'
        'To reproduce locally, please see: '
        'https://google.github.io/oss-fuzz/advanced-topics/reproducing'
        '#reproducing-build-failures\n\n'
        '<b>This bug tracker is not being monitored by OSS-Fuzz team.</b> '
        'If you have any questions, please create an issue at '
        'https://github.com/google/oss-fuzz/issues/new.\n\n'
        '**This bug will be automatically closed within a '
        'day once it is fixed.**', issue.body)

    self.assertTrue(issue.has_label('Proj-proj6'))
    self.assertTrue(issue.has_label('Type-Build-Failure'))

  def test_recovered_build_failure(self):
    """Test fixed build failures."""
    # Use the same status for all build types.
    self.mock.get.return_value = MockResponse(
        json.dumps({
            'projects': [{
                'history': [{
                    'finish_time': '2018-02-01T00:00:00.000000Z',
                    'build_id': 'proj0-id',
                    'success': True
                }],
                'name':
                    'proj0',
            }]
        }))

    data_types.OssFuzzBuildFailure(
        id='proj0',
        project_name='proj0',
        last_checked_timestamp=datetime.datetime(2018, 1, 31),
        issue_id='1',
        consecutive_failures=2,
        build_type='fuzzing').put()

    issue = Issue()
    issue.open = True
    issue.add_label('Type-Build-Failure')
    issue.add_label('Proj-proj2')
    issue.summary = 'Build failure in proj2'
    issue.body = 'Build failure'

    self.itm.issues[1] = issue

    self.app.get('/build-status')
    self.assertEqual(0, data_types.OssFuzzBuildFailure.query().count())

    issue = self.itm.issues[1]
    self.assertEqual('Verified', issue.status)
    self.assertEqual('The latest build has succeeded, closing this issue.',
                     issue.comment)

  def test_missing_builds(self):
    """Test missing builds."""

    def _mock_requests_get(url):
      """Mock fetch."""
      if url == oss_fuzz_build_status.FUZZING_STATUS_URL:
        return MockResponse(
            json.dumps({
                'projects': [
                    {
                        'history': [{
                            'finish_time': '2018-01-30T00:00:00.000000Z',
                            'build_id': 'proj0-id-f',
                            'success': True
                        }],
                        'name':
                            'proj0',
                    },
                    {
                        'history': [{
                            'finish_time': '2018-02-01T00:00:00.000000Z',
                            'build_id': 'proj0-id-f',
                            'success': True
                        }],
                        'name':
                            'proj1',
                    },
                ]
            }))

      assert url == oss_fuzz_build_status.COVERAGE_STATUS_URL
      return MockResponse(
          json.dumps({
              'projects': [
                  {
                      'history': [{
                          'finish_time': '2018-02-01T00:00:00.000000Z',
                          'build_id': 'proj0-id-c',
                          'success': True
                      }],
                      'name':
                          'proj0',
                  },
                  {
                      'history': [{
                          'finish_time': '2018-01-30T00:00:00.000000Z',
                          'build_id': 'proj1-id-c',
                          'success': True
                      }],
                      'name':
                          'proj1',
                  },
              ]
          }))

    self.mock.get.side_effect = _mock_requests_get
    self.app.get('/build-status')
    self.mock.log_error.assert_has_calls([
        mock.call('proj0 has not been built in fuzzing config for 2 days.'),
        mock.call('proj1 has not been built in coverage config for 2 days.')
    ])

  def test_disabled_project(self):
    """Test disabled project."""
    # Return the same status for all build types.
    self.mock.get.return_value = MockResponse(
        json.dumps({
            'projects': [{
                'history': [{
                    'finish_time': '2018-02-01T00:00:00.000000Z',
                    'build_id': 'proj2-id',
                    'success': False
                }],
                'name':
                    'disabled_proj',
            },]
        }))

    # Only fuzzing build type failure should be stored.
    data_types.OssFuzzBuildFailure(
        id='disabled_proj',
        project_name='disabled_proj',
        last_checked_timestamp=datetime.datetime(2018, 1, 31),
        consecutive_failures=1,
        build_type='fuzzing').put()

    self.app.get('/build-status')
    six.assertCountEqual(self, [
        {
            'build_type': 'fuzzing',
            'consecutive_failures': 1,
            'issue_id': None,
            'last_checked_timestamp': datetime.datetime(2018, 1, 31, 0, 0),
            'project_name': u'disabled_proj',
        },
    ], [
        failure.to_dict() for failure in data_types.OssFuzzBuildFailure.query()
    ])

    self.assertEqual(0, len(self.itm.issues))

  def test_reminder(self):
    """Test reminders."""
    # Return the same status for all build types.
    self.mock.get.return_value = MockResponse(
        json.dumps({
            'projects': [
                {
                    'history': [{
                        'finish_time': '2018-02-01T00:00:00.000000Z',
                        'build_id': 'proj0-id',
                        'success': False
                    }],
                    'name':
                        'proj0',
                },
                {
                    'history': [{
                        'finish_time': '2018-02-01T00:00:00.000000Z',
                        'build_id': 'proj0-id',
                        'success': False
                    }],
                    'name':
                        'proj1',
                },
            ]
        }))

    data_types.OssFuzzProject(
        id='proj0', name='proj0', ccs=['a@user.com']).put()
    data_types.OssFuzzBuildFailure(
        id='proj0',
        project_name='proj0',
        last_checked_timestamp=datetime.datetime(2018, 1, 31),
        issue_id='1',
        consecutive_failures=7,
        build_type='fuzzing').put()
    data_types.OssFuzzProject(
        id='proj1', name='proj1', ccs=['a@user.com']).put()
    data_types.OssFuzzBuildFailure(
        id='proj1',
        project_name='proj1',
        last_checked_timestamp=datetime.datetime(2018, 1, 31),
        issue_id='2',
        consecutive_failures=3,
        build_type='fuzzing').put()

    self.itm.issues[1] = Issue()
    self.itm.issues[2] = Issue()

    self.app.get('/build-status')
    self.assertEqual(
        'Friendly reminder that the the build is still failing.\n'
        'Please try to fix this failure to ensure that fuzzing remains '
        'productive.\n'
        'Latest build log: https://oss-fuzz-build-logs.storage.googleapis.com/'
        'log-proj0-id.txt\n', self.itm.issues[1].comment)
    self.assertEqual('', self.itm.issues[2].comment)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for performing remote tasks on linux."""
from fabric import api

from local.remote import utils
from local.remote.handlers import posix


class Handler(posix.Handler):
  """Handler for performing linux task."""

  def __init__(self, instance_name, project=None, zone=None):
    assert project, 'Need to specify a project via --project argument.'
    assert zone, 'Need to specify a zone via --zone argument.'

    super(Handler, self).__init__(
        instance_name=instance_name,
        platform='linux',
        project=project,
        zone=zone)

    # FIXME: Make these configurable.
    self.username = 'clusterfuzz'
    self.clusterfuzz_parent_path = '/mnt/scratch0'
    self.clusterfuzz_parent_path_outside = '/var/scratch0'

    (hostname, username,
     ssh_key_file_path) = utils.get_host_user_and_ssh_key_path(
         self.instance_name, self.project, self.zone)
    api.env.host_string = '{username}@{hostname}'.format(
        username=username, hostname=hostname)
    api.env.key_filename = ssh_key_file_path
    api.env.use_shell = True

  def _path_outside_docker(self, path):
    """Return the path outside docker."""
    return path.replace(self.clusterfuzz_parent_path,
                        self.clusterfuzz_parent_path_outside)

  def _run(self, command):
    """Custom _run that ensures that command is run inside docker container."""
    print('Running: ' + command)
    return api.sudo('docker exec {user} bash -c "{command}"'.format(
        user=self.username, command=command.replace('"', '\\"')))

  def _copy_staging_archive_from_local_to_remote(self, local_zip_path):
    """Copy staging archive from local to remote."""
    remote_zip_path = (
        '{clusterfuzz_parent_path}/{staging_source_filename}'.format(
            clusterfuzz_parent_path=self.clusterfuzz_parent_path,
            staging_source_filename=self.staging_source_filename))
    self._run('rm -f ' + remote_zip_path)

    api.sudo('chmod a+w ' + self.clusterfuzz_parent_path_outside)
    api.put(local_zip_path, self._path_outside_docker(remote_zip_path))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Download testcase tests."""
import unittest

import flask
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.testcase_detail import download_testcase
from libs import helpers


@test_utils.with_cloud_emulators('datastore')
class OAuthHandlerTest(unittest.TestCase):
  """Test OAuthHandler."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.blobs.get_blob_info',
        'libs.access.check_access_and_get_testcase',
        'libs.handler.get_email_and_access_token',
        'libs.gcs.get_signed_url',
    ])
    self.mock.get_email_and_access_token.return_value = ('test@test.com',
                                                         'access_token')
    self.mock.get_signed_url.return_value = 'https://SIGNED_URL'

    self.mock.get_blob_info.side_effect = (
        lambda key: storage.GcsBlobInfo('blobs-bucket', key, 'file.tar.gz', 1))

  def _test_download(self, handler_class):
    """Test download."""
    testcase = data_types.Testcase()
    testcase.minimized_keys = 'key'
    testcase.put()

    self.mock.check_access_and_get_testcase.return_value = testcase
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=handler_class.as_view('/'))
    app = webtest.TestApp(flaskapp)
    resp = app.get('/?id=%s' % testcase.key.id())

    self.assertEqual(302, resp.status_int)
    # pylint: disable=line-too-long
    # According to https://cloud.google.com/appengine/docs/go/how-requests-are-handled
    # Appengine replaces X-AppEngine-BlobKey with the blob content.
    self.assertEqual(
        'https://SIGNED_URL&response-content-disposition=attachment'
        '%3B+filename%3Dtestcase-1-file.tar.gz', resp.location)

  def test_download(self):
    """Test download on Handler."""
    self._test_download(download_testcase.Handler)


@test_utils.with_cloud_emulators('datastore')
class GetTestcaseBlobInfoTest(unittest.TestCase):
  """Test get_testcase_blob_info."""

  def setUp(self):
    self.testcase = data_types.Testcase()
    self.testcase.put()

    test_helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.blobs.get_blob_info',
    ])

    self.key = 'key'
    self.mock.get_blob_info.side_effect = (
        lambda key: storage.GcsBlobInfo('blobs-bucket', key, 'file.tar.gz', 1))

  def test_minimized(self):
    """Test getting minimized testcase."""
    self.testcase.minimized_keys = self.key
    self.testcase.fuzzed_keys = ''
    self.testcase.put()

    blob_info, using_minimized_keys = (
        download_testcase.get_testcase_blob_info(self.testcase))
    self.assertEqual(self.key, blob_info.key())
    self.assertEqual('file.tar.gz', blob_info.filename)
    self.assertTrue(using_minimized_keys)

  def test_unminimized(self):
    """Test getting unminimized."""
    self.testcase.minimized_keys = ''
    self.testcase.fuzzed_keys = self.key
    self.testcase.put()

    blob_info, using_minimized_keys = (
        download_testcase.get_testcase_blob_info(self.testcase))
    self.assertEqual(self.key, blob_info.key())
    self.assertEqual('file.tar.gz', blob_info.filename)
    self.assertFalse(using_minimized_keys)

  def test_error(self):
    """Test no blobkey."""
    self.testcase.minimized_keys = ''
    self.testcase.fuzzed_keys = ''
    self.testcase.put()

    with self.assertRaises(helpers.EarlyExitException) as cm:
      download_testcase.get_testcase_blob_info(self.testcase)

    self.assertEqual(400, cm.exception.status)
    self.assertEqual((
        "The testcase (%d) doesn't have fuzzed keys." % self.testcase.key.id()),
                     str(cm.exception))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzing engine interface."""

import os
import shutil

from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers import utils as fuzzer_utils
from clusterfuzz._internal.bot.fuzzers.syzkaller import constants
from clusterfuzz._internal.bot.fuzzers.syzkaller import runner
from clusterfuzz._internal.metrics import profiler
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell
from clusterfuzz.fuzz import engine

BIN_FOLDER_PATH = 'bin'
REPRO_TIME = 70
CORPUS_DB_FILENAME = 'corpus.db'


class SyzkallerError(Exception):
  """Base exception class."""


class SyzkallerOptions(engine.FuzzOptions):
  """Represents options passed to the engine. Can be overridden to provide more
  options."""

  def __init__(self, corpus_dir, arguments, strategies, fuzz_corpus_dirs,
               extra_env):
    super().__init__(corpus_dir, arguments, strategies)
    self.fuzz_corpus_dirs = fuzz_corpus_dirs
    self.extra_env = extra_env


class SyzkallerEngine(engine.Engine):
  """Syzkaller fuzzing engine implementation."""

  @property
  def name(self):
    return 'syzkaller'

  def prepare_binary_path(self):
    """Prepares the path for the syzkaller binary.

    Returns:
      The full path of the binary folder.
    """
    syzkaller_path = os.path.join(
        environment.get_value('BUILD_DIR'), 'syzkaller')
    if not os.path.exists(syzkaller_path):
      raise SyzkallerError('syzkaller not found in build')
    binary_folder = os.path.join(syzkaller_path, BIN_FOLDER_PATH)

    for root, _, filenames in os.walk(binary_folder):
      for filename in filenames:
        absolute_file_path = os.path.join(root, filename)
        os.chmod(absolute_file_path, 0o755)

    return binary_folder

  def prepare(self, corpus_dir, target_path, unused_build_dir):  # pylint: disable=unused-argument
    """Prepare for a fuzzing session, by generating options and making
    syzkaller binaries executable.

    Args:
      corpus_dir: The main corpus directory.
      target_path: Path to the target.
      build_dir: Path to the build directory.

    Returns:
      A FuzzOptions object."""
    self.prepare_binary_path()
    config = runner.get_config()
    return SyzkallerOptions(
        corpus_dir,
        config,
        strategies={},
        fuzz_corpus_dirs=None,
        extra_env=None)

  def _create_temp_corpus_dir(self, name):
    """Create temporary corpus directory."""
    new_corpus_directory = os.path.join(fuzzer_utils.get_temp_dir(), name)
    engine_common.recreate_directory(new_corpus_directory)
    return new_corpus_directory

  def _get_device_corpus_db_filename(self):
    """Return device-specific corpus db filename."""
    return environment.get_value('ANDROID_SERIAL') + '.db'

  def save_corpus(self, source_dir, destination_dir):
    """Saves syzkaller to folder so it is backed up to the cloud.

    Args:
      source_dir: Folder where syzkaller corpus is.
      destination_dir: Folder where the corpus is synced with the cloud.
    """
    source_file = os.path.join(source_dir, CORPUS_DB_FILENAME)
    shell.create_directory(destination_dir)
    destination_file = os.path.join(destination_dir,
                                    self._get_device_corpus_db_filename())
    if os.path.isfile(source_file) and (
        not os.path.exists(destination_file) or
        (os.path.getsize(source_file) > os.path.getsize(destination_file))):
      shutil.copy(source_file, destination_file)

  def init_corpus(self, source_dir, destination_dir):
    """Uses corpus from the cloud to initialize syzkaller corpus.

    Args:
      source_dir: Folder where the corpus is downloaded from the cloud.
      destination_dir: Folder where syzkaller will be looking for corpus.
    """
    source_file = os.path.join(source_dir,
                               self._get_device_corpus_db_filename())
    shell.create_directory(destination_dir)
    destination_file = os.path.join(destination_dir, CORPUS_DB_FILENAME)
    if os.path.isfile(source_file):
      shutil.copy(source_file, destination_file)

  def fuzz(self, target_path, options, unused_reproducers_dir=None, max_time=0):
    """Run a fuzz session.

    Args:
      target_path: Path to the target.
      options: The FuzzOptions object returned by prepare().
      reproducers_dir: The directory to put reproducers in when crashes
          are found.
      max_time: Maximum allowed time for the fuzzing to run.

    Returns:
      A FuzzResult object.
    """
    profiler.start_if_needed('syzkaller_kasan')
    syzkaller_runner = runner.get_runner(target_path)

    # Directory to place new units.
    self._create_temp_corpus_dir('new')

    args = options.arguments

    self.init_corpus(options.corpus_dir, runner.get_work_dir())
    fuzz_result = syzkaller_runner.fuzz(max_time, additional_args=args)
    self.save_corpus(runner.get_work_dir(), options.corpus_dir)
    return fuzz_result

  def reproduce(self, target_path, input_path, arguments, max_time):  # pylint: disable=unused-argument
    """Reproduce a crash given an input.
       Example: ./syz-crush -config my.cfg -infinite=false -restart_time=20s
        crash-qemu-1-1455745459265726910

    Args:
      target_path: Path to the target.
      input_path: Path to the reproducer input.
      arguments: Additional arguments needed for reproduction.
      max_time: Maximum allowed time for the reproduction.

    Returns:
      A ReproduceResult.
    """
    binary_dir = self.prepare_binary_path()
    syzkaller_runner = runner.get_runner(
        os.path.join(binary_dir, constants.SYZ_REPRO))
    repro_args = runner.get_config()
    repro_args.extend(
        ['-infinite=false', '-restart_time={}s'.format(REPRO_TIME), input_path])
    result = syzkaller_runner.repro(max_time, repro_args=repro_args)

    return engine.ReproduceResult(result.command, result.return_code,
                                  result.time_executed, result.output)

  def minimize_corpus(self, target_path, arguments, input_dirs, output_dir,
                      unused_reproducers_dir, unused_max_time):
    """Optional (but recommended): run corpus minimization.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for corpus minimization.
      input_dirs: Input corpora.
      output_dir: Output directory to place minimized corpus.
      reproducers_dir: The directory to put reproducers in when crashes are
          found.
      max_time: Maximum allowed time for the minimization.

    Returns:
      A FuzzResult object.
    """
    raise NotImplementedError

  def minimize_testcase(self, target_path, arguments, input_path, output_path,
                        max_time):
    """Optional (but recommended): Minimize a testcase.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for testcase minimization.
      input_path: Path to the reproducer input.
      output_path: Path to the minimized output.
      max_time: Maximum allowed time for the minimization.

    Returns:
      A ReproduceResult.
    """
    raise NotImplementedError

  def cleanse(self, target_path, arguments, input_path, output_path, max_time):
    """Optional (but recommended): Cleanse a testcase.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for testcase cleanse.
      input_path: Path to the reproducer input.
      output_path: Path to the cleansed output.
      max_time: Maximum allowed time for the cleanse.

    Returns:
      A ReproduceResult.
    """
    raise NotImplementedError



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Request specific caching.."""

import collections

import flask

from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.metrics import logs


def get_current_request():
  """Get the current request."""
  return flask.request


def get_cache_backing():
  """Get the cache backing for saving current context data."""
  return flask.g


class _FifoRequestCache(memoize.FifoInMemory):
  """In memory caching engine scoped to a request."""

  def __init__(self, cache_key, capacity):
    super(_FifoRequestCache, self).__init__(capacity)
    self._cache_key = str(cache_key)

  @property
  def cache(self):
    """Get the cache backing."""
    cache_backing = get_cache_backing()
    if not cache_backing:
      # Not a cache (e.g. in a unit test). Should not happen in production.
      logs.log_error('No container found for cache.')
      return None

    key = '__cache:' + self._cache_key

    backing = getattr(cache_backing, key, None)
    if backing is None:
      backing = collections.OrderedDict()
      setattr(cache_backing, key, backing)

    return backing


def wrap(capacity):
  """Wraps a function to use the per request cache."""

  def decorator(func):
    """Decorator function."""
    engine = _FifoRequestCache(id(func), capacity)
    return memoize.wrap(engine)(func)

  return decorator



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Android init scripts."""

from clusterfuzz._internal.bot.init_scripts import init_runner
from clusterfuzz._internal.platforms import android
from clusterfuzz._internal.system import environment

TIME_SINCE_REBOOT_MIN_THRESHOLD = 30 * 60  # 30 minutes.


def run():
  """Run Android initialization."""
  init_runner.run()

  # Set cuttlefish device serial if needed.
  if environment.is_android_cuttlefish():
    android.adb.set_cuttlefish_device_serial()

  # Check if we need to reflash device to latest build.
  android.flash.flash_to_latest_build_if_needed()

  # Reconnect to cuttlefish device if connection is ever lost.
  if environment.is_android_cuttlefish():
    android.adb.connect_to_cuttlefish_device()

  # Reboot to bring device in a good state if not done recently.
  if android.adb.time_since_last_reboot() > TIME_SINCE_REBOOT_MIN_THRESHOLD:
    android.device.reboot()

  # Make sure that device is in a good condition before we move forward.
  android.adb.wait_until_fully_booted()

  # Wait until battery charges to a minimum level and temperature threshold.
  android.battery.wait_until_good_state()

  # Initialize environment settings.
  android.device.initialize_environment()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run tests."""
import unittest

import mock

from clusterfuzz._internal.tests.test_libs import helpers
from python.bot.startup import run


class UpdateSourceCodeIfNeededTest(unittest.TestCase):
  """Test update_source_code_if_needed."""

  def setUp(self):
    helpers.patch(self, [
        'python.bot.startup.run.stop_heartbeat',
        'clusterfuzz._internal.bot.tasks.update_task.get_newer_source_revision',
        'clusterfuzz._internal.bot.tasks.update_task.update_source_code',
    ])

  def test_not_update(self):
    """Test newer source revision is None."""
    self.mock.get_newer_source_revision.return_value = None
    run.update_source_code_if_needed()
    self.assertEqual(0, self.mock.stop_heartbeat.call_count)
    self.assertEqual(0, self.mock.update_source_code.call_count)

  def test_update(self):
    """Test update source."""
    self.mock.get_newer_source_revision.return_value = 'revision'
    run.update_source_code_if_needed()
    self.mock.stop_heartbeat.assert_called_once_with()
    self.mock.update_source_code.assert_called_once_with()


class RunLoopTest(unittest.TestCase):
  """Test run_loop."""

  def setUp(self):
    helpers.patch(self, [
        'atexit.register',
        'python.bot.startup.run.start_bot',
        'python.bot.startup.run.start_heartbeat',
        'python.bot.startup.run.stop_heartbeat',
        'python.bot.startup.run.update_source_code_if_needed',
        'python.bot.startup.run.sleep',
        'clusterfuzz._internal.datastore.data_handler.bot_run_timed_out',
    ])

  def test_loop(self):
    """Test looping until break."""
    self.mock.bot_run_timed_out.side_effect = [False, False, True]
    self.mock.start_bot.return_value = 1

    run.run_loop('bot command', 'heartbeat command')

    self.assertEqual(3, self.mock.update_source_code_if_needed.call_count)
    self.assertEqual(3, self.mock.start_heartbeat.call_count)
    self.assertEqual(1, self.mock.register.call_count)
    self.assertEqual(0, self.mock.stop_heartbeat.call_count)  # Handled at exit.
    self.assertEqual(3, self.mock.start_bot.call_count)
    self.assertEqual(3, self.mock.bot_run_timed_out.call_count)
    self.assertEqual(2, self.mock.sleep.call_count)

    self.mock.update_source_code_if_needed.assert_has_calls(
        [mock.call(), mock.call(), mock.call()])
    self.mock.start_bot.assert_has_calls([
        mock.call('bot command'),
        mock.call('bot command'),
        mock.call('bot command'),
    ])
    self.mock.bot_run_timed_out.assert_has_calls(
        [mock.call(), mock.call(), mock.call()])
    self.mock.sleep.assert_has_calls([mock.call(3), mock.call(3)])
